"""Test for parallel build."""

import os

import pytest
from sphinx.util.parallel import parallel_available


@pytest.mark.skipif(
    not parallel_available, reason="sphinx parallel builds not available"
)
@pytest.mark.sphinx("html", testroot="parallel")
def test_parallel(make_app, app_params) -> None:
    args, kwargs = app_params
    app0 = make_app(*args, **kwargs)
    app0.parallel = 4
    app0.build()
    assert not app0._warning.getvalue()
    # update files to trigger merge of citations domain data
    docnames = {
        "{0}/doc{1:02d}".format(folder, i)
        for i in (1, 3, 5, 7, 11, 15, 19)
        for folder in ("regular", "foot")
    }
    for docname in docnames:
        rstname = app0.env.doc2path(docname)
        htmlname = app0.builder.get_outfilename(docname)
        mtime = (app0.builder.outdir / htmlname).stat().st_mtime
        os.utime(app0.srcdir / rstname, (mtime + 5, mtime + 5))
    updated = set(app0.builder.get_outdated_docs())
    assert updated == docnames
    app1 = make_app(*args, **kwargs)
    app1._warning.seek(0)
    app1._warning.truncate()
    app1.build()
    assert not app1._warning.getvalue()



"""
    test_autodoc
    ~~~~~~~~~~~~

    Test with autodoc.
"""

import sphinx

from test.common import (
    html_citation_refs_single,
    html_citations,
    html_footnote_refs,
    html_footnotes,
)

import pytest


@pytest.mark.skipif(
    sphinx.version_info < (7, 0),
    reason="autoapi appears broken on sphinx < 7.0",
)
@pytest.mark.sphinx("html", testroot="autoapi")
def test_autoapi(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    root = (app.outdir / "index.html").read_text()
    output = (app.outdir / "autoapi/some_module/cite/index.html").read_text()
    labels = [
        "One",
        "Two",
        "Thr",
        "Fou",
        "Fiv",
        "Six",
        "Sev",
        "Eig",
        "Nin",
        "Ten",
        "Ele",
    ]
    titles = [
        "Een",
        "Twee",
        "Drie",
        "Vier",
        "Vijf",
        "Zes",
        "Zeven",
        "Acht",
        "Negen",
        "Tien",
        "Elf",
    ]
    for label, title in zip(labels, titles):
        assert len(html_citation_refs_single(label=label).findall(output)) == 1
        assert len(html_citations(label=label).findall(root)) == 1
        match_ref = html_citation_refs_single(label=label).search(output)
        match = html_citations(label=label).search(root)
        assert match_ref
        assert match
        assert match_ref.group("refid") == match.group("id_")
        assert title in match.group("text")
        # no backrefs as citations are in other document
        # assert match_ref.group("id_") == match.group("backref")
    output2 = (app.outdir / "autoapi/some_module/footcite/index.html").read_text()
    assert len(html_footnote_refs().findall(output2)) == 11
    for title in titles:
        text = ".*" + title + ".*"
        assert len(html_footnotes(text=text).findall(output2)) == 1
        match = html_footnotes(text=text).search(output2)
        assert match
        id_ = match.group("id_")
        assert len(html_footnote_refs(refid=id_).findall(output2)) == 1



extensions = ["sphinxcontrib.bibtex", "autoapi.extension"]
exclude_patterns = ["_build"]
bibtex_bibfiles = ["test.bib"]
autoapi_python_class_content = "both"  # document __init__ method too
autoapi_dirs = ["some_module/"]
autoapi_keep_files = True  # useful for debugging
autoapi_add_toctree_entry = False



"""Citation keys."""

import re
from typing import Iterable, NamedTuple


class CitationTarget(NamedTuple):
    """Citation key, pre-text, and post-text."""

    key: str
    pre: str
    post: str


_re_citation_target = re.compile(
    r"\s*([{](?P<pre>[^{}]+)[}])?"
    r"\s*(?P<key>[^{}\s,]+)"
    r"\s*([{](?P<post>[^{}]+)[}])?\s*"
)


def parse_citation_targets(targets: str, pos=0) -> Iterable[CitationTarget]:
    """Parse citation target string into a list of citation keys."""
    match = _re_citation_target.match(targets, pos=pos)
    if match is None:
        raise ValueError(f"malformed citation target: {targets}")
    yield CitationTarget(
        key=match.group("key") or "",
        pre=match.group("pre") or "",
        post=match.group("post") or "",
    )
    end = match.end()
    if end < len(targets):
        if targets[end] != ",":
            raise ValueError(f"malformed citation target: {targets}")
        yield from parse_citation_targets(targets, end + 1)



"""
.. autofunction:: sphinxcontrib.bibtex.style.template.join(\
        sep='', sep2=None, last_sep=None, other=None)

.. autofunction:: sphinxcontrib.bibtex.style.template.sentence(\
        capfirst=False, capitalize=False, add_period=True, \
        sep=', ', sep2=None, last_sep=None, other=None)

.. autofunction:: sphinxcontrib.bibtex.style.template.names(\
        role, sep='', sep2=None, last_sep=None, other=None)

.. autofunction:: sphinxcontrib.bibtex.style.template.entry_label()

.. autofunction:: sphinxcontrib.bibtex.style.template.reference()

.. autofunction:: sphinxcontrib.bibtex.style.template.footnote_reference()
"""

from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, cast

import docutils.nodes
import pybtex_docutils
from pybtex.richtext import Text
from pybtex.style.template import (
    FieldIsMissing,
    Node,
    _format_list,
    field,
    first_of,
    optional,
    tag,
)
from sphinx.util.nodes import make_refnode

from sphinxcontrib.bibtex.nodes import raw_latex
from sphinxcontrib.bibtex.richtext import BaseReferenceText

if TYPE_CHECKING:
    from pybtex.backends import BaseBackend
    from pybtex.richtext import BaseText
    from pybtex.style import FormattedEntry
    from sphinx.builders import Builder


# extended from pybtex: also copies the docstring into the wrapped object
def node(f):
    n = Node(f.__name__, f)
    n.__doc__ = f.__doc__
    return n


# copied from pybtex join but extended to allow "et al" formatting
@node
def join(children, data, sep="", sep2=None, last_sep=None, other=None):
    """Join text fragments together."""

    if sep2 is None:
        sep2 = sep
    if last_sep is None:
        last_sep = sep
    parts = [part for part in _format_list(children, data) if part]
    if len(parts) <= 1:
        return Text(*parts)
    elif len(parts) == 2:
        return Text(sep2).join(parts)
    elif other is None:
        return Text(last_sep).join([Text(sep).join(parts[:-1]), parts[-1]])
    else:
        return Text(parts[0], other)


@node
def join2(children, data, sep1="", sep2=""):
    """Join text fragments together."""
    if not children:
        return Text()
    else:
        return join(sep=sep1)[children[0], join(sep=sep2)[children[1:]]].format_data(
            data
        )


# copied from pybtex names but using the new join
@node
def sentence(
    children,
    data,
    capfirst=False,
    capitalize=False,
    add_period=True,
    sep=", ",
    sep2=None,
    last_sep=None,
    other=None,
):
    """Join text fragments, capitalize the first letter,
    and add a period to the end.
    """
    text = join(sep=sep, sep2=sep2, last_sep=last_sep, other=other)[
        children
    ].format_data(data)
    if capfirst:
        text = text.capfirst()
    if capitalize:
        text = text.capitalize()
    if add_period:
        text = text.add_period()
    return text


# copied from pybtex names but using the new join allowing "et al" formatting
@node
def names(children, data, role, **kwargs):
    """Return formatted names."""
    assert not children
    try:
        persons = data["entry"].persons[role]
    except KeyError:
        raise FieldIsMissing(role, data["entry"])
    style = data["style"]
    formatted_names = [
        style.person.style_plugin.format(person, style.person.abbreviate)
        for person in persons
    ]
    return join(**kwargs)[formatted_names].format_data(data)


@node
def entry_label(children, data) -> "BaseText":
    """Node for inserting the label of a formatted entry."""
    assert not children
    entry = cast("FormattedEntry", data["formatted_entry"])
    return Text(entry.label)


class SphinxReferenceInfo(NamedTuple):
    """Tuple containing reference info to enable sphinx to resolve a reference
    to a citation.
    """

    builder: "Builder"  #: The Sphinx builder.
    fromdocname: str  #: Document name of the citation reference.
    todocname: str  #: Document name of the bibliography.
    citation_id: str  #: Unique id of the citation within the bibliography.
    title: str  #: Title attribute for reference node.
    pre_text: str  #: Text to come before citation.
    post_text: str  #: Text to come after citation.


class SphinxReferenceText(BaseReferenceText[SphinxReferenceInfo]):
    """Pybtex rich text class generating
    a docutils reference node to a citation
    for use with :class:`SphinxReferenceInfo`.
    """

    def render(self, backend: "BaseBackend") -> List[docutils.nodes.Element]:
        assert isinstance(
            backend, pybtex_docutils.Backend
        ), "SphinxReferenceText only supports the docutils backend"
        info = self.info[0]
        if info.builder.name == "latex":
            key = f"cite.{info.todocname}:{info.citation_id}"
            return (
                [raw_latex(f"\\hyperlink{{{key}}}{{")]
                + super().render(backend)
                + [raw_latex("}")]
            )
        elif info.builder.name == "rinoh":
            children = super().render(backend)
            refid = f"%{info.todocname}#{info.citation_id}"
            refnode = docutils.nodes.citation_reference(
                text=children[0], refid=refid, reftitle=info.title
            )
            refnode.extend(children[1:])
            return [refnode]
        else:
            children = super().render(backend)
            # make_refnode only takes a single child
            refnode2 = make_refnode(
                builder=info.builder,
                fromdocname=info.fromdocname,
                todocname=info.todocname,
                targetid=info.citation_id,
                child=children[0],
                title=info.title,
            )
            refnode2.extend(children[1:])  # type: ignore
            return [refnode2]


@node
def reference(children, data: Dict[str, Any]):
    """Pybtex node for inserting a docutils reference node to a citation.
    The children of the node
    comprise the content of the reference, and any referencing information
    is stored in the *reference_info* key of the *data*.
    The data must also contain a *style* key pointing to the corresponding
    :class:`~sphinxcontrib.bibtex.style.referencing.BaseReferenceStyle`.
    """
    parts = _format_list(children, data)
    info = data["reference_info"]
    assert isinstance(info, SphinxReferenceInfo)
    return SphinxReferenceText(info, *parts)


@node
def pre_text(children, data: Dict[str, Any]):
    assert not children
    info = data["reference_info"]
    assert isinstance(info, SphinxReferenceInfo)
    return Text(info.pre_text)


@node
def post_text(children, data: Dict[str, Any]):
    assert not children
    info = data["reference_info"]
    assert isinstance(info, SphinxReferenceInfo)
    return Text(info.post_text)


class FootReferenceInfo(NamedTuple):
    """Tuple containing reference info to enable sphinx to resolve a footnote
    reference.
    """

    key: str  #: Citation key.
    document: "docutils.nodes.document"  #: Current docutils document.
    refname: str  #: Citation reference name.


class FootReferenceText(BaseReferenceText[FootReferenceInfo]):
    """Pybtex rich text class generating
    a docutils footnote_reference node to a citation
    for use with :class:`FootReferenceInfo`.
    """

    def render(self, backend: "BaseBackend"):
        assert isinstance(
            backend, pybtex_docutils.Backend
        ), "FootReferenceText only supports the docutils backend"
        info = self.info[0]
        # see docutils.parsers.rst.states.Body.footnote_reference()
        refnode = docutils.nodes.footnote_reference(
            "[#%s]_" % info.key, refname=info.refname, auto=1
        )
        info.document.note_autofootnote_ref(refnode)
        info.document.note_footnote_ref(refnode)
        return [refnode]


@node
def footnote_reference(children, data: Dict[str, Any]):
    """Pybtex node for inserting a footnote_reference docutils node.
    Any referencing information
    is stored in the *reference_info* key of the *data*.
    The data must also contain a *style* key pointing to the corresponding
    :class:`~sphinxcontrib.bibtex.style.referencing.BaseReferenceStyle`.
    """
    assert not children
    info = data["reference_info"]
    assert isinstance(info, FootReferenceInfo)
    # we need to give the footnote text some fake content
    # otherwise pybtex richtext engine will mess things up
    return FootReferenceText(info, "#")


@node
def year(children, data: Dict[str, Any]) -> "BaseText":
    assert not children
    return first_of[optional[field("year")], "n.d."].format_data(data)


@node
def author_or_editor_or_title(children, data, **kwargs):
    assert not children
    return first_of[
        optional[names("author", **kwargs)],
        optional[names("editor", **kwargs)],
        tag("em")[field("title")],
    ].format_data(data)



from pybtex.plugin import register_plugin
from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.labels import BaseLabelStyle

extensions = ["sphinxcontrib.bibtex"]
exclude_patterns = ["_build"]
bibtex_bibfiles = ["test.bib"]
bibtex_default_style = "mystyle"


# a simple label style which uses the bibtex keys for labels
class MyLabelStyle(BaseLabelStyle):
    def format_labels(self, sorted_entries):
        for entry in sorted_entries:
            yield entry.key


class MyStyle(UnsrtStyle):
    default_label_style = MyLabelStyle


register_plugin("pybtex.style.formatting", "mystyle", MyStyle)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, List

from sphinxcontrib.bibtex.style.template import reference

from . import BaseReferenceStyle, BracketStyle, PersonStyle

if TYPE_CHECKING:
    from pybtex.richtext import BaseText
    from pybtex.style.template import Node


@dataclass
class ExtraAuthorReferenceStyle(BaseReferenceStyle):
    """Reference just by author names."""

    #: Bracket style.
    bracket: BracketStyle = field(default_factory=BracketStyle)

    #: Person style.
    person: PersonStyle = field(default_factory=PersonStyle)

    def role_names(self) -> Iterable[str]:
        return [
            f"{capfirst}author{parenthetical}{full_author}"
            for parenthetical in ["par", ""]
            for capfirst in (["", "c"] if parenthetical == "" else [""])
            for full_author in ["", "s"]
        ]

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        return self.bracket.outer(
            children,
            brackets="par" in role_name,
            capfirst="c" in role_name,
        )

    def inner(self, role_name: str) -> "Node":
        return reference[self.person.author_or_editor_or_title(full="s" in role_name)]



from sphinxcontrib.bibtex.style.template import join, join2, sentence


def test_join() -> None:
    assert str(join.format()) == ""
    assert str(join["a", "b", "c", "d", "e"].format()) == "abcde"
    join_sep = join(sep=", ", sep2=" and ", last_sep=", and ")
    assert str(join_sep["Tom", "Jerry"].format()) == "Tom and Jerry"
    assert (
        str(join_sep["Billy", "Willy", "Dilly"].format()) == "Billy, Willy, and Dilly"
    )
    join_other = join(other=" et al.")
    assert str(join_other["Billy", "Willy", "Dilly"].format()) == "Billy et al."


def test_join2() -> None:
    assert str(join2.format()) == ""
    assert str(join2["a", "b", "c", "d", "e"].format()) == "abcde"
    join_sep = join2(sep1=";", sep2=",")
    assert str(join_sep["Tom", "Jerry"].format()) == "Tom;Jerry"
    assert str(join_sep["Billy", "Willy", "Dilly"].format()) == "Billy;Willy,Dilly"


def test_sentence() -> None:
    assert str(sentence.format()) == ""
    assert (
        str(
            sentence(capitalize=True, sep=" ")[
                "mary", "had", "a", "little", "lamb"
            ].format()
        )
        == "Mary had a little lamb."
    )
    assert (
        str(sentence(capitalize=False, add_period=False)["uno", "dos", "tres"].format())
        == "uno, dos, tres"
    )
    assert (
        str(sentence(capfirst=True, other=" and more")["uno", "dos", "tres"].format())
        == "Uno and more."
    )



from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable, List

from sphinxcontrib.bibtex.style.template import join

from . import BaseReferenceStyle

if TYPE_CHECKING:
    from pybtex.richtext import BaseText
    from pybtex.style.template import Node


@dataclass
class ExtraEmptyReferenceStyle(BaseReferenceStyle):
    """A style which generates nothing, similar to LaTeX's nocite."""

    def role_names(self) -> Iterable[str]:
        return ["empty"]

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        return join

    def inner(self, role_name: str) -> "Node":
        return join



import re
from test.common import (
    html_citation_refs,
    html_citations,
    html_footnote_refs,
    html_footnotes,
)
from typing import Set

import pytest


def citation_refs(output) -> Set[str]:
    return {match.group("label") for match in html_citation_refs().finditer(output)}


def citations(output) -> Set[str]:
    return {match.group("label") for match in html_citations().finditer(output)}


@pytest.mark.sphinx("html", testroot="bibliography_empty", freshenv=True)
def test_bibliography_empty(app, warning) -> None:
    app.build()
    assert not warning.getvalue()


@pytest.mark.sphinx("html", testroot="bibliography_header")
def test_bibliography_header(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8")
    assert "Regular Citations" in output
    assert "Footnote Citations" in output


@pytest.mark.sphinx(
    "pseudoxml",
    testroot="bibliography_empty",
    freshenv=True,
    confoverrides={
        "bibtex_bibliography_header": ".. rubric:: Regular Citations",
        "bibtex_footbibliography_header": ".. rubric:: Footnote Citations",
    },
)
def test_bibliography_empty_no_header(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.pseudoxml").read_text(encoding="utf-8")
    assert "Regular Citations" not in output
    assert "Footnote Citations" not in output
    assert "<rubric" not in output


@pytest.mark.sphinx("html", testroot="bibliography_style_default")
def test_bibliography_style_default(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    # ensure Man09 is cited with plain style and not with alpha style
    assert citation_refs(output) == citations(output) == {"1"}


@pytest.mark.sphinx("html", testroot="bibliography_style_label_1")
def test_bibliography_style_label_1(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    # the custom style uses keys as labels
    # citations
    assert len(html_citations(label="myfancybibtexkey").findall(output)) == 1
    assert len(html_citations(label="myotherfancybibtexkey").findall(output)) == 1
    assert len(html_citation_refs(label="myfancybibtexkey").findall(output)) == 1
    assert len(html_citation_refs(label="myotherfancybibtexkey").findall(output)) == 1


@pytest.mark.sphinx("html", testroot="bibliography_style_label_2")
def test_bibliography_style_label_2(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    assert len(html_citation_refs(label="APAa").findall(output)) == 1
    assert len(html_citation_refs(label="APAb").findall(output)) == 1


@pytest.mark.sphinx("html", testroot="bibliography_style_nowebref")
def test_bibliography_style_nowebref(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8")
    # the custom style suppresses web links
    assert "http://arxiv.org" not in output
    assert "http://dx.doi.org" not in output


@pytest.mark.sphinx("html", testroot="bibliography_bad_option")
def test_bibliography_bad_option(app, warning) -> None:
    app.build()
    assert 'unknown option: "thisisintentionallyinvalid"' in warning.getvalue()


# see issue 87
@pytest.mark.sphinx("html", testroot="bibliography_key_prefix")
def test_bibliography_key_prefix(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "doc0.html").read_text()
    assert citations(output) == citation_refs(output) == {"AMan09", "AEve03"}
    output = (app.outdir / "doc1.html").read_text()
    assert citations(output) == citation_refs(output) == {"BMan09"}


@pytest.mark.sphinx("html", testroot="bibliography_label_prefix_1")
def test_bibliography_label_prefix_1(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "doc1.html").read_text()
    assert citations(output) == citation_refs(output) == {"A1"}
    output = (app.outdir / "doc2.html").read_text()
    assert citations(output) == citation_refs(output) == {"B1"}


@pytest.mark.sphinx("html", testroot="bibliography_label_prefix_2")
def test_bibliography_label_prefix_2(app, warning) -> None:
    doc1_refs = {"AFM12", "ABlu83", "AGIH02", "AWS14"}
    doc1_cites = {"ABlu83", "AFM12", "AGIH02", "AWS14"}
    doc2_refs = {"BShi13"}
    doc2_cites = {"BShi13"}
    sum_refs = {"CMcMahonKM10", "CRMM11", "CRM09", "CMM03", "CHdJMD13", "AFM12"}
    sum_cites = {"CMcMahonKM10", "CRMM11", "CRM09", "CMM03", "CHdJMD13"}
    app.build()
    assert not warning.getvalue()
    output1 = (app.outdir / "doc1.html").read_text(encoding="utf-8-sig")
    # use <= instead of == as there are some extra reference nodes
    assert doc1_refs <= citation_refs(output1)
    assert doc1_cites == citations(output1)
    output2 = (app.outdir / "doc2.html").read_text(encoding="utf-8-sig")
    assert doc2_refs <= citation_refs(output2)
    assert doc2_cites == citations(output2)
    output3 = (app.outdir / "summary.html").read_text(encoding="utf-8-sig")
    assert sum_refs <= citation_refs(output3)
    assert sum_cites == citations(output3)
    # check citation reference from summary to doc1
    match1 = html_citations(label="AFM12").search(output1)
    match3 = html_citation_refs(label="AFM12").search(output3)
    assert match1
    assert match3
    assert match1.group("id_") == match3.group("refid")
    assert match3.group("refdoc") == "doc1.html"


# test order of bibliography entries when using an unsorted style
@pytest.mark.sphinx("html", testroot="bibliography_order_unsorted")
def test_bibliography_order_unsorted(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    assert re.search(
        r"<p>A\.\sFirst\. Test 1\.</p>.*<p>B\.\sSecond\. Test 2\.</p>",
        output,
        re.DOTALL,
    )


# see issue 187
@pytest.mark.sphinx("html", testroot="bibliography_multi_foot")
def test_bibliography_multi_foot(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8")
    assert output.count('<p class="rubric"') == 3
    assert len(re.findall(html_footnotes(id_="footcite-2009-mandel"), output)) == 1
    assert len(re.findall(html_footnotes(id_="footcite-2003-evensen"), output)) == 1
    assert len(re.findall(html_footnotes(id_="footcite-1986-lorenc"), output)) == 1
    assert (
        len(re.findall(html_footnote_refs(refid="footcite-2009-mandel"), output)) == 2
    )
    assert (
        len(re.findall(html_footnote_refs(refid="footcite-2003-evensen"), output)) == 1
    )
    assert (
        len(re.findall(html_footnote_refs(refid="footcite-1986-lorenc"), output)) == 1
    )


@pytest.mark.sphinx("html", testroot="bibliography_missing_field")
def test_bibliography_missing_field(app, warning) -> None:
    app.build()
    assert "missing year in testkey" in warning.getvalue()


@pytest.mark.sphinx("html", testroot="bibliography_content")
def test_bibliography_content(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output1 = (app.outdir / "doc1.html").read_text()
    output2 = (app.outdir / "doc2.html").read_text()
    output3 = (app.outdir / "doc3.html").read_text()
    assert citation_refs(output1) == {"One", "Two"}
    assert citations(output1) == {"One", "Two", "Thr"}
    assert citation_refs(output2) == {"Fou", "Fiv"}
    assert citations(output2) == {"Fiv", "Six"}
    assert not citation_refs(output3)
    assert citations(output3) == {"Fou"}


@pytest.mark.sphinx("html", testroot="bibliography_bad_key")
def test_bibliography_bad_key(app, warning) -> None:
    app.build()
    assert 'could not find bibtex key "badkey"' in warning.getvalue()


def url(link: str) -> str:
    return f'<a class="reference external" href="{link}">{link}</a>'


@pytest.mark.sphinx("html", testroot="bibliography_url")
def test_bibliography_url(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8")
    match1 = html_citations(label="Een").search(output)
    match2 = html_citations(label="Twe").search(output)
    match3 = html_citations(label="Dri").search(output)
    match4 = html_citations(label="Vie").search(output)
    assert match1 is not None
    assert match2 is not None
    assert match3 is not None
    assert match4 is not None
    assert url("https://github.com/") in match1.group("text")
    assert "aaa" + url("https://google.com/") + "bbb" in match2.group("text")
    assert url("https://youtube.com/") in match3.group("text")
    assert "URL: " + url("https://wikipedia.org/") in match4.group("text")
    # check tooltips do not contain \url command
    match4 = html_citation_refs(label="Een").search(output)
    match5 = html_citation_refs(label="Twe").search(output)
    match6 = html_citation_refs(label="Dri").search(output)
    match7 = html_citation_refs(label="Vie").search(output)
    assert match4 is not None
    assert match5 is not None
    assert match6 is not None
    assert match7 is not None
    assert "\\url " not in match4.group("title")
    assert "\\url " not in match5.group("title")
    assert "\\url " not in match6.group("title")
    assert "\\url " not in match7.group("title")


@pytest.mark.sphinx("html", testroot="bibliography_custom_ids")
def test_bibliography_custom_ids(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8")
    assert ' id="bibliography-id-1"' in output
    assert ' id="bibliography-id-2"' in output
    assert ' id="footbibliography-id-1"' in output
    assert ' id="footbibliography-id-2"' in output
    match1 = html_citations(text=".*Evensen.*").search(output)
    match2 = html_citations(text=".*Mandel.*").search(output)
    match3 = html_citations(text=".*Lorenc.*").search(output)
    assert match1 is not None
    assert match2 is not None
    assert match3 is not None
    assert match1.group("id_") == "cite-id-1-2003-evensen"
    assert match2.group("id_") == "cite-id-1-2009-mandel"
    assert match3.group("id_") == "cite-id-2-1986-lorenc"
    match1 = html_footnotes(text=".*Evensen.*").search(output)
    match2 = html_footnotes(text=".*Mandel.*").search(output)
    match3 = html_footnotes(text=".*Lorenc.*").search(output)
    assert match1 is not None
    assert match2 is not None
    assert match3 is not None
    assert match1.group("id_") == "footcite-id-2003-evensen"
    assert match2.group("id_") == "footcite-id-2009-mandel"
    assert match3.group("id_") == "footcite-id-1986-lorenc"



"""
    .. autoclass:: BibliographyTransform
        :show-inheritance:

        .. autoattribute:: default_priority
        .. automethod:: run
"""
from itertools import zip_longest
from typing import TYPE_CHECKING, cast

import docutils.nodes
from pybtex.plugin import find_plugin
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.logging import getLogger

from .directives import BibliographyKey
from .nodes import bibliography as bibliography_node

if TYPE_CHECKING:
    from sphinx.environment import BuildEnvironment

    from .domain import BibtexDomain

logger = getLogger(__name__)


def node_text_transform(node: docutils.nodes.Element) -> None:
    """Apply extra text transformations to a node."""
    for child, next_child in zip_longest(node.children[:], node.children[1:]):
        if isinstance(child, docutils.nodes.Text):
            if child.endswith(r"\url ") and isinstance(next_child, docutils.nodes.Text):
                node.replace(child, docutils.nodes.Text(child[:-5]))
                ref_node = docutils.nodes.reference(refuri=next_child.astext())
                ref_node += next_child
                node.replace(next_child, ref_node)
        elif isinstance(child, docutils.nodes.Element):  # pragma: no branch
            node_text_transform(child)


class BibliographyTransform(SphinxPostTransform):
    """A docutils transform to generate citation entries for
    bibliography nodes.
    """

    # transform must be applied before sphinx runs its ReferencesResolver
    # which has priority 10, so when ReferencesResolver calls the cite domain
    # resolve_xref, the target is present and all will work fine
    default_priority = 5
    backend = find_plugin("pybtex.backends", "docutils")()

    def run(self, **kwargs):
        """Transform each
        :class:`~sphinxcontrib.bibtex.nodes.bibliography` node into a
        list of citations.
        """
        env = cast("BuildEnvironment", self.document.settings.env)
        domain = cast("BibtexDomain", env.get_domain("cite"))
        # Can just use "findall" once docutils 0.18+ is required
        meth = "findall" if hasattr(self.document, "findall") else "traverse"
        for bibnode in getattr(self.document, meth)(bibliography_node):
            # reminder: env.docname may be equal to 'index' instead of
            # bibnode['docname'] in post-transform phase (e.g. latex builder)
            bib_key = BibliographyKey(docname=bibnode["docname"], id_=bibnode["ids"][0])
            bibliography = domain.bibliographies[bib_key]
            citations = [
                citation
                for citation in domain.citations
                if citation.bibliography_key == bib_key
            ]
            # create citation nodes for all references
            if bibliography.list_ == "enumerated":
                nodes = docutils.nodes.enumerated_list()
                nodes["enumtype"] = bibliography.enumtype
                if bibliography.start >= 1:
                    nodes["start"] = bibliography.start
                    env.temp_data["bibtex_enum_count"] = bibliography.start
                else:
                    nodes["start"] = env.temp_data.setdefault("bibtex_enum_count", 1)
            elif bibliography.list_ == "bullet":
                nodes = docutils.nodes.bullet_list()
            else:  # "citation"
                nodes = []
            for citation in citations:
                citation_node = bibliography.citation_nodes[citation.key]
                if bibliography.list_ in {"enumerated", "bullet"}:
                    citation_node += self.backend.paragraph(citation.formatted_entry)
                else:  # "citation"
                    # backrefs only supported in same document
                    backrefs = [
                        citation_ref.citation_ref_id
                        for citation_ref in domain.citation_refs
                        if bib_key.docname == citation_ref.docname
                        and citation.key
                        in {target.key for target in citation_ref.targets}
                    ]
                    if backrefs:
                        citation_node["backrefs"] = backrefs
                    citation_node += docutils.nodes.label(
                        "", citation.formatted_entry.label, support_smartquotes=False
                    )
                    citation_node += self.backend.paragraph(citation.formatted_entry)
                citation_node["docname"] = bib_key.docname
                node_text_transform(citation_node)
                nodes.append(citation_node)
                if bibliography.list_ == "enumerated":
                    env.temp_data["bibtex_enum_count"] += 1
            if citations:
                final_node = domain.bibliography_header.deepcopy()
                final_node += nodes
                bibnode.replace_self(final_node)
            else:
                bibnode.replace_self(docutils.nodes.target())



extensions = ["sphinxcontrib.bibtex"]
exclude_patterns = ["_build"]
bibtex_bibfiles = ["test.bib"]
bibtex_cite_id = "cite-id-{bibliography_count}-{key}"
bibtex_footcite_id = "footcite-id-{key}"
bibtex_bibliography_id = "bibliography-id-{bibliography_count}"
bibtex_footbibliography_id = "footbibliography-id-{footbibliography_count}"



import re
from dataclasses import dataclass, field
from test.common import html_citation_refs, html_citations, html_docutils_citation_refs
from typing import cast

import pybtex.plugin
import pytest
from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.template import words

import sphinxcontrib.bibtex.plugin
from sphinxcontrib.bibtex.domain import BibtexDomain
from sphinxcontrib.bibtex.style.referencing import BracketStyle, PersonStyle
from sphinxcontrib.bibtex.style.referencing.author_year import AuthorYearReferenceStyle


@pytest.mark.sphinx("html", testroot="citation_not_found")
def test_citation_not_found(app, warning) -> None:
    app.build()
    assert 'could not find bibtex key "nosuchkey1"' in warning.getvalue()
    assert 'could not find bibtex key "nosuchkey2"' in warning.getvalue()


# test mixing of ``:cite:`` and ``[]_`` (issue 2)
@pytest.mark.sphinx("html", testroot="citation_mixed")
def test_citation_mixed(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    domain = cast(BibtexDomain, app.env.get_domain("cite"))
    assert len(domain.citation_refs) == 1
    citation_ref = domain.citation_refs.pop()
    assert citation_ref.targets == [("Test", "", "")]
    assert citation_ref.docname == "adoc1"
    assert len(domain.citations) == 1
    citation = domain.citations.pop()
    assert citation.formatted_entry.label == "1"


@pytest.mark.sphinx("html", testroot="citation_multiple_keys")
def test_citation_multiple_keys(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    cits = {match.group("label") for match in html_citations().finditer(output)}
    citrefs = {match.group("label") for match in html_citation_refs().finditer(output)}
    assert {"App", "Bra"} == cits == citrefs


@pytest.mark.sphinx("html", testroot="citation_any_role")
def test_citation_any_role(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    cits = {match.group("label") for match in html_citations().finditer(output)}
    citrefs = {match.group("label") for match in html_citation_refs().finditer(output)}
    assert {"App", "Bra"} == cits == citrefs


def find_label(output: str, label: str):
    assert html_citation_refs(label=label).search(output) is not None


# see issue 85
@pytest.mark.sphinx("html", testroot="citation_no_author_no_key")
def test_citation_no_author_no_key(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    find_label(output, "<em>Software projects built on Mesos</em>, 2015")
    find_label(output, "2015")
    find_label(output, "Mandel, 2009")
    find_label(output, "2009")
    find_label(output, "<em>This citation only has a title</em>, n.d.")
    find_label(output, "n.d.")
    find_label(output, "Whatever, 2021")
    find_label(output, "2021")


# test cites spanning multiple lines (issue 205)
@pytest.mark.sphinx("html", testroot="citation_whitespace")
def test_citation_whitespace(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    # ensure Man09 is cited
    assert len(html_citation_refs(label="Fir").findall(output)) == 1
    assert len(html_citation_refs(label="Sec").findall(output)) == 1


# test document not in toctree (issue 228)
@pytest.mark.sphinx("pseudoxml", testroot="citation_from_orphan")
def test_citation_from_orphan(app, warning) -> None:
    app.build()
    assert not warning.getvalue()


@pytest.mark.sphinx("text", testroot="citation_roles")
def test_citation_roles_label(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    tests = [
        ("p", " [dDEF03] "),
        ("ps", " [dDEF03] "),
        ("t", " de Du *et al.* [dDEF03] "),
        ("ts", " de Du, Em, and Fa [dDEF03] "),
        ("ct", " De Du *et al.* [dDEF03] "),
        ("cts", " De Du, Em, and Fa [dDEF03] "),
        ("labelpar", " [dDEF03] "),
        ("label", " dDEF03 "),
        ("yearpar", " [2003] "),
        ("year", " 2003 "),
        ("authorpar", " [de Du *et al.*] "),
        ("authorpars", " [de Du, Em, and Fa] "),
        ("author", " de Du *et al.* "),
        ("authors", " de Du, Em, and Fa "),
        ("cauthor", " De Du *et al.* "),
        ("cauthors", " De Du, Em, and Fa "),
        ("empty", " AAA  AAA "),
        ("p", " [aA01, BC02] "),
        ("ps", " [aA01, BC02] "),
        ("t", " al Ap [aA01], Be and Ci [BC02] "),
        ("ts", " al Ap [aA01], Be and Ci [BC02] "),
        ("ct", " Al Ap [aA01], Be and Ci [BC02] "),
        ("cts", " Al Ap [aA01], Be and Ci [BC02] "),
        ("labelpar", " [aA01, BC02] "),
        ("label", " aA01, BC02 "),
        ("yearpar", " [2001, 2002] "),
        ("year", " 2001, 2002 "),
        ("authorpar", " [al Ap, Be and Ci] "),
        ("authorpars", " [al Ap, Be and Ci] "),
        ("author", " al Ap, Be and Ci "),
        ("authors", " al Ap, Be and Ci "),
        ("cauthor", " Al Ap, Be and Ci "),
        ("cauthors", " Al Ap, Be and Ci "),
        ("empty", " BBB  BBB "),
        ("p", " [Ge04, Hu05, Ix06] "),
        ("ps", " [Ge04, Hu05, Ix06] "),
        ("t", " Ge [Ge04], Hu [Hu05], Ix [Ix06] "),
        ("ts", " Ge [Ge04], Hu [Hu05], Ix [Ix06] "),
        ("ct", " Ge [Ge04], Hu [Hu05], Ix [Ix06] "),
        ("cts", " Ge [Ge04], Hu [Hu05], Ix [Ix06] "),
        ("labelpar", " [Ge04, Hu05, Ix06] "),
        ("label", " Ge04, Hu05, Ix06 "),
        ("yearpar", " [2004, 2005, 2006] "),
        ("year", " 2004, 2005, 2006 "),
        ("authorpar", " [Ge, Hu, Ix] "),
        ("authorpars", " [Ge, Hu, Ix] "),
        ("author", " Ge, Hu, Ix "),
        ("authors", " Ge, Hu, Ix "),
        ("cauthor", " Ge, Hu, Ix "),
        ("cauthors", " Ge, Hu, Ix "),
        ("empty", " CCC  CCC "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":cite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None
    # check :cite:empty: generates citation
    assert "[Ju07] Jo Ju. Testseven. 2007." in output


@pytest.mark.sphinx(
    "text",
    testroot="citation_roles",
    confoverrides={"bibtex_reference_style": "author_year"},
)
def test_citation_roles_authoryear(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    tests = [
        ("p", " [de Du *et al.*, 2003] "),
        ("ps", " [de Du, Em, and Fa, 2003] "),
        ("alp", " de Du *et al.*, 2003 "),
        ("alps", " de Du, Em, and Fa, 2003 "),
        ("t", " de Du *et al.* [2003] "),
        ("ts", " de Du, Em, and Fa [2003] "),
        ("ct", " De Du *et al.* [2003] "),
        ("cts", " De Du, Em, and Fa [2003] "),
        ("labelpar", " [dDEF03] "),
        ("label", " dDEF03 "),
        ("yearpar", " [2003] "),
        ("year", " 2003 "),
        ("authorpar", " [de Du *et al.*] "),
        ("authorpars", " [de Du, Em, and Fa] "),
        ("author", " de Du *et al.* "),
        ("authors", " de Du, Em, and Fa "),
        ("cauthor", " De Du *et al.* "),
        ("cauthors", " De Du, Em, and Fa "),
        ("empty", " AAA  AAA "),
        ("p", " [al Ap, 2001, Be and Ci, 2002] "),
        ("ps", " [al Ap, 2001, Be and Ci, 2002] "),
        ("alp", " al Ap, 2001, Be and Ci, 2002 "),
        ("alps", " al Ap, 2001, Be and Ci, 2002 "),
        ("t", " al Ap [2001], Be and Ci [2002] "),
        ("ts", " al Ap [2001], Be and Ci [2002] "),
        ("ct", " Al Ap [2001], Be and Ci [2002] "),
        ("cts", " Al Ap [2001], Be and Ci [2002] "),
        ("labelpar", " [aA01, BC02] "),
        ("label", " aA01, BC02 "),
        ("yearpar", " [2001, 2002] "),
        ("year", " 2001, 2002 "),
        ("authorpar", " [al Ap, Be and Ci] "),
        ("authorpars", " [al Ap, Be and Ci] "),
        ("author", " al Ap, Be and Ci "),
        ("authors", " al Ap, Be and Ci "),
        ("cauthor", " Al Ap, Be and Ci "),
        ("cauthors", " Al Ap, Be and Ci "),
        ("empty", " BBB  BBB "),
        ("p", " [Ge, 2004, Hu, 2005, Ix, 2006] "),
        ("ps", " [Ge, 2004, Hu, 2005, Ix, 2006] "),
        ("alp", " Ge, 2004, Hu, 2005, Ix, 2006 "),
        ("alps", " Ge, 2004, Hu, 2005, Ix, 2006 "),
        ("t", " Ge [2004], Hu [2005], Ix [2006] "),
        ("ts", " Ge [2004], Hu [2005], Ix [2006] "),
        ("ct", " Ge [2004], Hu [2005], Ix [2006] "),
        ("cts", " Ge [2004], Hu [2005], Ix [2006] "),
        ("labelpar", " [Ge04, Hu05, Ix06] "),
        ("label", " Ge04, Hu05, Ix06 "),
        ("yearpar", " [2004, 2005, 2006] "),
        ("year", " 2004, 2005, 2006 "),
        ("authorpar", " [Ge, Hu, Ix] "),
        ("authorpars", " [Ge, Hu, Ix] "),
        ("author", " Ge, Hu, Ix "),
        ("authors", " Ge, Hu, Ix "),
        ("cauthor", " Ge, Hu, Ix "),
        ("cauthors", " Ge, Hu, Ix "),
        ("empty", " CCC  CCC "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":cite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None
    # check :cite:empty: generates citation
    assert "[Ju07] Jo Ju. Testseven. 2007." in output


@pytest.mark.sphinx("text", testroot="citation_roles_pre_post")
def test_citation_roles_label_pre_post(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    lab1 = "dDEF03"
    lab2 = "aA01"
    lab3 = "BC02"
    ref1 = "de Du *et al.*"
    ref2 = "al Ap"
    ref3 = "Be and Ci"
    tests = [
        ("p", f" A [see {lab1}] "),
        ("p", f" B [{lab1}, p. 1] "),
        ("p", f" C [see {lab1}, p. 1] "),
        ("t", f" A {ref1} [see {lab1}] "),
        ("t", f" B {ref1} [{lab1}, p. 1] "),
        ("t", f" C {ref1} [see {lab1}, p. 1] "),
        ("p", f" A [see {lab2}, {lab3}] "),
        ("p", f" B [{lab2}, p. 1, {lab3}] "),
        ("p", f" C [see {lab2}, p. 1, {lab3}] "),
        ("p", f" D [{lab2}, see {lab3}] "),
        ("p", f" E [{lab2}, {lab3}, p. 2] "),
        ("p", f" F [{lab2}, see {lab3}, p. 2] "),
        ("p", f" G [see {lab2}, see {lab3}] "),
        ("p", f" H [{lab2}, p. 1, {lab3}, p. 2] "),
        ("p", f" I [see {lab2}, {lab3}, p. 2] "),
        ("p", f" J [see {lab2}, p. 1, see {lab3}, p. 2] "),
        ("t", f" A {ref2} [see {lab2}], {ref3} [{lab3}] "),
        ("t", f" B {ref2} [{lab2}, p. 1], {ref3} [{lab3}] "),
        ("t", f" C {ref2} [see {lab2}, p. 1], {ref3} [{lab3}] "),
        ("t", f" D {ref2} [{lab2}], {ref3} [see {lab3}] "),
        ("t", f" E {ref2} [{lab2}], {ref3} [{lab3}, p. 2] "),
        ("t", f" F {ref2} [{lab2}], {ref3} [see {lab3}, p. 2] "),
        ("t", f" G {ref2} [see {lab2}], {ref3} [see {lab3}] "),
        ("t", f" H {ref2} [{lab2}, p. 1], {ref3} [{lab3}, p. 2] "),
        ("t", f" I {ref2} [see {lab2}], {ref3} [{lab3}, p. 2] "),
        ("t", f" J {ref2} [see {lab2}, p. 1], {ref3} [see {lab3}, p. 2] "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":cite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None, text


@pytest.mark.sphinx(
    "text",
    testroot="citation_roles_pre_post",
    confoverrides={"bibtex_reference_style": "author_year"},
)
def test_citation_roles_authoryear_pre_post(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    ref1 = "de Du *et al.*"
    ref2 = "al Ap"
    ref3 = "Be and Ci"
    tests = [
        ("p", f" A [see {ref1}, 2003] "),
        ("p", f" B [{ref1}, 2003, p. 1] "),
        ("p", f" C [see {ref1}, 2003, p. 1] "),
        ("t", f" A {ref1} [see 2003] "),
        ("t", f" B {ref1} [2003, p. 1] "),
        ("t", f" C {ref1} [see 2003, p. 1] "),
        ("p", f" A [see {ref2}, 2001, {ref3}, 2002] "),
        ("p", f" B [{ref2}, 2001, p. 1, {ref3}, 2002] "),
        ("p", f" C [see {ref2}, 2001, p. 1, {ref3}, 2002] "),
        ("p", f" D [{ref2}, 2001, see {ref3}, 2002] "),
        ("p", f" E [{ref2}, 2001, {ref3}, 2002, p. 2] "),
        ("p", f" F [{ref2}, 2001, see {ref3}, 2002, p. 2] "),
        ("p", f" G [see {ref2}, 2001, see {ref3}, 2002] "),
        ("p", f" H [{ref2}, 2001, p. 1, {ref3}, 2002, p. 2] "),
        ("p", f" I [see {ref2}, 2001, {ref3}, 2002, p. 2] "),
        ("p", f" J [see {ref2}, 2001, p. 1, see {ref3}, 2002, p. 2] "),
        ("t", f" A {ref2} [see 2001], {ref3} [2002] "),
        ("t", f" B {ref2} [2001, p. 1], {ref3} [2002] "),
        ("t", f" C {ref2} [see 2001, p. 1], {ref3} [2002] "),
        ("t", f" D {ref2} [2001], {ref3} [see 2002] "),
        ("t", f" E {ref2} [2001], {ref3} [2002, p. 2] "),
        ("t", f" F {ref2} [2001], {ref3} [see 2002, p. 2] "),
        ("t", f" G {ref2} [see 2001], {ref3} [see 2002] "),
        ("t", f" H {ref2} [2001, p. 1], {ref3} [2002, p. 2] "),
        ("t", f" I {ref2} [see 2001], {ref3} [2002, p. 2] "),
        ("t", f" J {ref2} [see 2001, p. 1], {ref3} [see 2002, p. 2] "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":cite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None, text


@pytest.mark.sphinx(
    "html",
    testroot="citation_roles",
    confoverrides={"bibtex_default_style": "plain", "bibtex_reference_style": "super"},
)
def test_citation_roles_super(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    # just a cursory check that superscript references are present
    assert '<sup><a class="reference internal" href="#' in output


@pytest.mark.sphinx(
    "pseudoxml",
    testroot="debug_bibtex_citation",
    confoverrides={"bibtex_reference_style": "non_existing"},
)
def test_citation_style_invalid(make_app, app_params) -> None:
    args, kwargs = app_params
    with pytest.raises(ImportError, match="plugin .*non_existing not found"):
        make_app(*args, **kwargs)


def my_bracket() -> BracketStyle:
    return BracketStyle(
        left="(",
        right=")",
        sep="; ",
        sep2="; ",
        last_sep="; ",
    )


def my_person() -> PersonStyle:
    return PersonStyle(
        style="last",
        abbreviate=False,
        sep=" & ",
        sep2=None,
        last_sep=None,
        other=" et al",
    )


@dataclass
class CustomReferenceStyle(AuthorYearReferenceStyle):
    bracket_textual: BracketStyle = field(default_factory=my_bracket)
    bracket_parenthetical: BracketStyle = field(default_factory=my_bracket)
    bracket_author: BracketStyle = field(default_factory=my_bracket)
    bracket_label: BracketStyle = field(default_factory=my_bracket)
    bracket_year: BracketStyle = field(default_factory=my_bracket)
    person: PersonStyle = field(default_factory=my_person)
    author_year_sep: str = " "


sphinxcontrib.bibtex.plugin.register_plugin(
    "sphinxcontrib.bibtex.style.referencing", "xxx_custom_xxx", CustomReferenceStyle
)


@pytest.mark.sphinx(
    "text",
    testroot="citation_roles",
    confoverrides={"bibtex_reference_style": "xxx_custom_xxx"},
)
def test_citation_style_custom(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    tests = [
        ("p", " (de Du et al 2003) "),
        ("ps", " (de Du & Em & Fa 2003) "),
        ("t", " de Du et al (2003) "),
        ("ts", " de Du & Em & Fa (2003) "),
        ("ct", " De Du et al (2003) "),
        ("cts", " De Du & Em & Fa (2003) "),
        ("labelpar", " (dDEF03) "),
        ("label", " dDEF03 "),
        ("yearpar", " (2003) "),
        ("year", " 2003 "),
        ("authorpar", " (de Du et al) "),
        ("authorpars", " (de Du & Em & Fa) "),
        ("author", " de Du et al "),
        ("authors", " de Du & Em & Fa "),
        ("cauthor", " De Du et al "),
        ("cauthors", " De Du & Em & Fa "),
        ("empty", " AAA  AAA "),
        ("p", " (al Ap 2001; Be & Ci 2002) "),
        ("ps", " (al Ap 2001; Be & Ci 2002) "),
        ("t", " al Ap (2001); Be & Ci (2002) "),
        ("ts", " al Ap (2001); Be & Ci (2002) "),
        ("ct", " Al Ap (2001); Be & Ci (2002) "),
        ("cts", " Al Ap (2001); Be & Ci (2002) "),
        ("labelpar", " (aA01; BC02) "),
        ("label", " aA01; BC02 "),
        ("yearpar", " (2001; 2002) "),
        ("year", " 2001; 2002 "),
        ("authorpar", " (al Ap; Be & Ci) "),
        ("authorpars", " (al Ap; Be & Ci) "),
        ("author", " al Ap; Be & Ci "),
        ("authors", " al Ap; Be & Ci "),
        ("cauthor", " Al Ap; Be & Ci "),
        ("cauthors", " Al Ap; Be & Ci "),
        ("empty", " BBB  BBB "),
        ("p", " (Ge 2004; Hu 2005; Ix 2006) "),
        ("ps", " (Ge 2004; Hu 2005; Ix 2006) "),
        ("t", " Ge (2004); Hu (2005); Ix (2006) "),
        ("ts", " Ge (2004); Hu (2005); Ix (2006) "),
        ("ct", " Ge (2004); Hu (2005); Ix (2006) "),
        ("cts", " Ge (2004); Hu (2005); Ix (2006) "),
        ("labelpar", " (Ge04; Hu05; Ix06) "),
        ("label", " Ge04; Hu05; Ix06 "),
        ("yearpar", " (2004; 2005; 2006) "),
        ("year", " 2004; 2005; 2006 "),
        ("authorpar", " (Ge; Hu; Ix) "),
        ("authorpars", " (Ge; Hu; Ix) "),
        ("author", " Ge; Hu; Ix "),
        ("authors", " Ge; Hu; Ix "),
        ("cauthor", " Ge; Hu; Ix "),
        ("cauthors", " Ge; Hu; Ix "),
        ("empty", " CCC  CCC "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":cite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None
    # check :cite:empty: generates citation
    assert "[Ju07] Jo Ju. Testseven. 2007." in output


@pytest.mark.sphinx("text", testroot="citation_style_round_brackets")
def test_citation_style_round_brackets(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    assert "(Evensen, 2003)" in output
    assert "Evensen (2003)" in output


@pytest.mark.sphinx("html", testroot="citation_toctree")
def test_citation_toctree(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    output1 = (app.outdir / "adoc1.html").read_text(encoding="utf-8-sig")
    output2 = (app.outdir / "adoc2.html").read_text(encoding="utf-8-sig")
    assert len(html_citations(label="1").findall(output)) == 1
    assert len(html_citation_refs(label="1").findall(output1)) == 1
    assert len(html_docutils_citation_refs(label=r"Test2").findall(output2)) == 1
    assert len(html_citations(label="Test2").findall(output2)) == 1


@pytest.mark.sphinx("html", testroot="debug_bibtex_citation")
def test_citation_tooltip(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    assert len(html_citations(label="tes").findall(output)) == 1
    assert (
        len(html_citation_refs(label="tes", title=r"The title\.").findall(output)) == 1
    )


@pytest.mark.sphinx(
    "html", testroot="debug_bibtex_citation", confoverrides={"bibtex_tooltips": False}
)
def test_citation_tooltip2(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    assert len(html_citations(label="tes").findall(output)) == 1
    assert len(html_citation_refs(label="tes", title=None).findall(output)) == 1


class CustomTooltipStyle(UnsrtStyle):
    def get_misc_template(self, e):
        return words["whoop whoop"]


pybtex.plugin.register_plugin(
    "pybtex.style.formatting", "xxx_custom_tooltip_xxx", CustomTooltipStyle
)


@pytest.mark.sphinx(
    "html",
    testroot="debug_bibtex_citation",
    confoverrides={"bibtex_tooltips_style": "xxx_custom_tooltip_xxx"},
)
def test_citation_tooltip3(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    assert len(html_citations(label="tes").findall(output)) == 1
    assert (
        len(html_citation_refs(label="tes", title="whoop whoop").findall(output)) == 1
    )


# https://bitbucket.org/pybtex-devs/pybtex/issues/442/author-names-with-special-latex-characters
@pytest.mark.xfail(reason="upstream bug")
@pytest.mark.sphinx("html", testroot="citation_label_special_chars")
def test_citation_label_special_chars(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text(encoding="utf-8-sig")
    assert len(html_citations(label="SBV09").findall(output)) == 1



from pybtex.style.names import BaseNameStyle, name_part
from pybtex.style.template import join


class LastNameStyle(BaseNameStyle):
    """A simple name style for formatting the last name of an author."""

    def format(self, person, abbr=True):
        """Format last names."""

        return join[
            name_part(tie=True)[person.rich_prelast_names],
            name_part[person.rich_last_names],
        ]



"""
    .. autoclass:: FootBibliographyDirective

        .. automethod:: run
"""

from typing import TYPE_CHECKING, cast

from docutils.parsers.rst import Directive

from .bibfile import _make_ids

if TYPE_CHECKING:
    from sphinx.environment import BuildEnvironment

    from .domain import BibtexDomain
    from .foot_domain import BibtexFootDomain


class FootBibliographyDirective(Directive):

    """Class for processing the :rst:dir:`footbibliography` directive."""

    required_arguments = 0
    optional_arguments = 0
    has_content = False

    def run(self):
        """Set file dependencies, and insert the footnotes that were created
        earlier by :meth:`.foot_roles.FootCiteRole.result_nodes`.
        """
        env = cast("BuildEnvironment", self.state.document.settings.env)
        foot_old_refs = env.temp_data.setdefault("bibtex_foot_old_refs", set())
        foot_new_refs = env.temp_data.setdefault("bibtex_foot_new_refs", set())
        footbibliography_count = env.temp_data["bibtex_footbibliography_count"] = (
            env.temp_data.get("bibtex_footbibliography_count", 0) + 1
        )
        if not foot_new_refs:
            return []
        else:
            foot_old_refs |= foot_new_refs
            foot_new_refs.clear()
            # bibliography stored in env.temp_data["bibtex_foot_bibliography"]
            foot_domain = cast("BibtexFootDomain", env.get_domain("footcite"))
            foot_bibliography, env.temp_data["bibtex_foot_bibliography"] = (
                env.temp_data["bibtex_foot_bibliography"],
                foot_domain.bibliography_header.deepcopy(),
            )
            domain = cast("BibtexDomain", env.get_domain("cite"))
            for bibfile in domain.bibdata.bibfiles:
                env.note_dependency(bibfile)
            foot_bibliography["ids"] += _make_ids(
                docname=env.docname,
                lineno=self.lineno,
                ids=set(self.state.document.ids.keys()),
                raw_id=env.app.config.bibtex_footbibliography_id.format(
                    footbibliography_count=footbibliography_count
                ),
            )
            self.state.document.note_explicit_target(
                foot_bibliography, foot_bibliography
            )
            return [foot_bibliography]



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, List, Union

from sphinxcontrib.bibtex.style.template import footnote_reference, join

from . import BaseReferenceStyle, BracketStyle, PersonStyle

if TYPE_CHECKING:
    from pybtex.richtext import BaseText
    from pybtex.style.template import Node


@dataclass
class BasicFootParentheticalReferenceStyle(BaseReferenceStyle):
    """Parenthetical footnote reference."""

    def role_names(self) -> Iterable[str]:
        return ["p", "ps"]

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        return join[children]

    def inner(self, role_name: str) -> "Node":
        return footnote_reference


@dataclass
class BasicFootTextualReferenceStyle(BaseReferenceStyle):
    """Textual footnote reference."""

    #: Bracket style.
    bracket: BracketStyle = field(default_factory=BracketStyle)

    #: Person style.
    person: PersonStyle = field(default_factory=PersonStyle)

    #: Separator between text and reference.
    text_reference_sep: Union["BaseText", str] = " "

    def role_names(self) -> Iterable[str]:
        return [
            f"{capfirst}t{full_author}"
            for capfirst in ["", "c"]
            for full_author in ["", "s"]
        ]

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        return self.bracket.outer(children, brackets=False, capfirst="c" in role_name)

    def inner(self, role_name: str) -> "Node":
        return join(sep=self.text_reference_sep)[
            self.person.author_or_editor_or_title(full="s" in role_name),
            footnote_reference,
        ]



import pytest


@pytest.mark.sphinx("html", testroot="spurious_div")
def test_spurious_div(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    assert output.count("<div") == output.count("</div>")


@pytest.mark.sphinx("html", testroot="spurious_div_2")
def test_spurious_div_2(app) -> None:
    # note: build will generate warning, is expected, not checking
    app.build()
    output = (app.outdir / "index.html").read_text()
    assert output.count("<div") == output.count("</div>")



import pytest


@pytest.mark.sphinx("html", testroot="filter")
def test_filter(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    assert "Tralalala" in output
    assert "ideetje" not in output
    assert "Jakkamakka" not in output
    output = (app.outdir / "or.html").read_text()
    assert "Tralalala" not in output
    assert "ideetje" in output
    assert "Jakkamakka" in output
    output = (app.outdir / "noteq.html").read_text()
    assert "Tralalala" in output
    assert "ideetje" in output
    assert "Jakkamakka" not in output
    output = (app.outdir / "lt.html").read_text()
    assert "Tralalala" in output
    assert "ideetje" not in output
    assert "Jakkamakka" not in output
    output = (app.outdir / "lte.html").read_text()
    assert "Tralalala" in output
    assert "ideetje" not in output
    assert "Jakkamakka" in output
    output = (app.outdir / "gt.html").read_text()
    assert "Tralalala" not in output
    assert "ideetje" in output
    assert "Jakkamakka" not in output
    output = (app.outdir / "gte.html").read_text()
    assert "Tralalala" not in output
    assert "ideetje" in output
    assert "Jakkamakka" in output
    output = (app.outdir / "key.html").read_text()
    assert "Tralalala" not in output
    assert "ideetje" in output
    assert "Jakkamakka" not in output
    output = (app.outdir / "false.html").read_text()
    assert "Tralalala" not in output
    assert "ideetje" not in output
    assert "Jakkamakka" not in output
    output = (app.outdir / "true.html").read_text()
    assert "Tralalala" in output
    assert "ideetje" in output
    assert "Jakkamakka" in output
    output = (app.outdir / "title.html").read_text()
    assert "Tralalala" not in output
    assert "ideetje" not in output
    assert "Jakkamakka" in output


@pytest.mark.sphinx("html", testroot="filter_fix_author_keyerror")
def test_filter_fix_author_keyerror(app) -> None:
    app.build()


@pytest.mark.sphinx("html", testroot="filter_option_clash")
def test_filter_option_clash(app, warning) -> None:
    app.build()
    warnings = warning.getvalue()
    assert ":filter: overrides :all:" in warnings
    assert ":filter: overrides :cited:" in warnings
    assert ":filter: overrides :notcited:" in warnings


@pytest.mark.sphinx("html", testroot="filter_syntax_error")
def test_filter_syntax_error(app, warning) -> None:
    app.build()
    assert warning.getvalue().count("syntax error in :filter: expression") == 9



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Union

from sphinxcontrib.bibtex.style.referencing import (
    BracketStyle,
    GroupReferenceStyle,
    PersonStyle,
)

from .basic_label import (
    BasicLabelParentheticalReferenceStyle,
    BasicLabelTextualReferenceStyle,
)
from .extra_author import ExtraAuthorReferenceStyle
from .extra_empty import ExtraEmptyReferenceStyle
from .extra_label import ExtraLabelReferenceStyle
from .extra_year import ExtraYearReferenceStyle

if TYPE_CHECKING:
    from pybtex.richtext import BaseText


@dataclass
class LabelReferenceStyle(GroupReferenceStyle):
    """Textual or parenthetical reference by label,
    or just by author, label, or year.
    """

    #: Bracket style for textual citations (:cite:t: and variations).
    bracket_textual: BracketStyle = field(default_factory=BracketStyle)

    #: Bracket style for parenthetical citations
    #: (:cite:p: and variations).
    bracket_parenthetical: BracketStyle = field(default_factory=BracketStyle)

    #: Bracket style for author citations
    #: (:cite:author: and variations).
    bracket_author: BracketStyle = field(default_factory=BracketStyle)

    #: Bracket style for label citations
    #: (:cite:label: and variations).
    bracket_label: BracketStyle = field(default_factory=BracketStyle)

    #: Bracket style for year citations
    #: (:cite:year: and variations).
    bracket_year: BracketStyle = field(default_factory=BracketStyle)

    #: Person style.
    person: PersonStyle = field(default_factory=PersonStyle)

    #: Separator between text and reference for textual citations.
    text_reference_sep: Union["BaseText", str] = " "

    #: Separator between pre-text and citation.
    pre_text_sep: Union["BaseText", str] = " "

    #: Separator between citation and post-text.
    post_text_sep: Union["BaseText", str] = ", "

    def __post_init__(self):
        self.styles.extend(
            [
                BasicLabelParentheticalReferenceStyle(
                    bracket=self.bracket_parenthetical,
                    pre_text_sep=self.pre_text_sep,
                    post_text_sep=self.post_text_sep,
                ),
                BasicLabelTextualReferenceStyle(
                    bracket=self.bracket_textual,
                    person=self.person,
                    text_reference_sep=self.text_reference_sep,
                    pre_text_sep=self.pre_text_sep,
                    post_text_sep=self.post_text_sep,
                ),
                ExtraAuthorReferenceStyle(
                    bracket=self.bracket_author, person=self.person
                ),
                ExtraLabelReferenceStyle(bracket=self.bracket_label),
                ExtraYearReferenceStyle(bracket=self.bracket_year),
                ExtraEmptyReferenceStyle(),
            ]
        )
        super().__post_init__()



"""
    Classes and methods to work with bib files.

    .. autoclass:: BibFile
        :members:

    .. autoclass:: BibData
        :members:

    .. autofunction:: normpath_filename

    .. autofunction:: parse_bibdata

    .. autofunction:: is_bibdata_outdated

    .. autofunction:: process_bibdata
"""
import math
import os.path
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Set

from docutils.nodes import make_id
from pybtex.database import BibliographyData, BibliographyDataError
from pybtex.database.input.bibtex import Parser
from sphinx.util.logging import getLogger

if TYPE_CHECKING:
    from sphinx.environment import BuildEnvironment


logger = getLogger(__name__)


class BibFile(NamedTuple):
    """Contains information about a parsed bib file."""

    mtime: float  #: Modification time of file when last parsed.
    keys: Dict[str, None]  #: Set of keys for this bib file as ordered dict.


class BibData(NamedTuple):
    """Contains information about a collection of bib files."""

    encoding: str  #: Encoding of all bib files.
    bibfiles: Dict[str, BibFile]  #: Maps bib filename to information about it.
    data: BibliographyData  #: Data parsed from all bib files.


def normpath_filename(env: "BuildEnvironment", filename: str) -> str:
    """Return normalised path to *filename* for the given environment *env*."""
    return os.path.normpath(env.relfn2path(filename.strip())[1])


def get_mtime(bibfilename: str) -> float:
    try:
        return os.path.getmtime(bibfilename)
    except OSError:
        return -math.inf


def parse_bibdata(bibfilenames: List[str], encoding: str) -> BibData:
    """Parse *bibfilenames* with given *encoding*, and return parsed data."""
    parser = Parser(encoding)
    bibfiles: Dict[str, BibFile] = {}
    keys: Dict[str, None] = {}
    for filename in bibfilenames:
        logger.info("parsing bibtex file {0}... ".format(filename), nonl=True)
        if not os.path.isfile(filename):
            logger.warning(
                "could not open bibtex file {0}.".format(filename),
                type="bibtex",
                subtype="bibfile_error",
            )
            new_keys: Dict[str, None] = {}
        else:
            try:
                parser.parse_file(filename)
            except BibliographyDataError as exc:
                logger.warning(
                    "bibliography data error in {0}: {1}".format(filename, exc),
                    type="bibtex",
                    subtype="bibfile_data_error",
                )
            keys, old_keys = dict.fromkeys(parser.data.entries.keys()), keys
            assert all(key in keys for key in old_keys)
            new_keys = dict.fromkeys(key for key in keys if key not in old_keys)
            logger.info("parsed {0} entries".format(len(new_keys)))
        bibfiles[filename] = BibFile(mtime=get_mtime(filename), keys=new_keys)
    return BibData(encoding=encoding, bibfiles=bibfiles, data=parser.data)


def is_bibdata_outdated(
    bibdata: BibData, bibfilenames: List[str], encoding: str
) -> bool:
    return (
        bibdata.encoding != encoding
        or list(bibdata.bibfiles) != bibfilenames
        or any(
            bibfile.mtime != get_mtime(filename)
            for filename, bibfile in bibdata.bibfiles.items()
        )
    )


def process_bibdata(
    bibdata: BibData, bibfilenames: List[str], encoding: str
) -> BibData:
    """Parse *bibfilenames* and store parsed data in *bibdata*."""
    logger.info("checking bibtex cache... ", nonl=True)
    if is_bibdata_outdated(bibdata, bibfilenames, encoding):
        logger.info("out of date")
        return parse_bibdata(bibfilenames, encoding)
    else:
        logger.info("up to date")
        return bibdata


# function does not really fit in any module, but used by both
# cite and footcite domains, so for now it's residing here
def _make_ids(docname: str, lineno: int, ids: Set[str], raw_id: str) -> List[str]:
    if raw_id:
        id_ = make_id(raw_id)
        if id_ in ids:
            logger.warning(
                f"duplicate citation id {id_}",
                location=(docname, lineno),
                type="bibtex",
                subtype="duplicate_id",
            )
            return []
        else:
            ids.add(id_)
            return [id_]
    else:
        return []



from abc import ABC
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple, Union

import pybtex.plugin
from pybtex.richtext import Tag, Text

from sphinxcontrib.bibtex.style.template import (
    author_or_editor_or_title,
    join,
    names,
    sentence,
)

if TYPE_CHECKING:
    from pybtex.database import Entry
    from pybtex.richtext import BaseText
    from pybtex.style import FormattedEntry
    from pybtex.style.names import BaseNameStyle
    from pybtex.style.template import Node

    from sphinxcontrib.bibtex.richtext import ReferenceInfo


@dataclass
class BaseReferenceStyle(ABC):
    """Base class for citation reference styles.

    For consistency, all subclasses of this class must be decorated
    as a :class:`dataclass`,
    and must provide a type annotation and default value for all attributes
    (unless ``init=False`` is used, in which case they can be
    initialized in :meth:`~dataclass.__post_init__`).
    This allows client code to instantiate any reference style
    without needing to specify any arguments through the constructor.
    """

    # see https://stackoverflow.com/a/59987363 as to why this is here
    def __post_init__(self):
        pass

    def role_names(self) -> Iterable[str]:
        """Get list of role names supported by this style."""
        raise NotImplementedError

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        """Returns outer template for formatting the references."""
        raise NotImplementedError

    def inner(self, role_name: str) -> "Node":
        """Returns inner template for formatting the references."""
        raise NotImplementedError


def format_references(
    style: BaseReferenceStyle,
    role_name: str,
    references: Iterable[Tuple["Entry", "FormattedEntry", "ReferenceInfo"]],
) -> "BaseText":
    """Format the list of references according to the given role.

    First formats each reference using the style's
    :meth:`~BaseReferenceStyle.inner` method,
    then joins all these formatted references together using
    the style's :meth:`~BaseReferenceStyle.outer` method.
    """
    children = [
        style.inner(role_name).format_data(
            data=dict(
                entry=entry,
                formatted_entry=formatted_entry,
                reference_info=info,
                style=style,
            )
        )
        for entry, formatted_entry, info in references
    ]
    return style.outer(role_name, children).format()


@dataclass
class BracketStyle:
    """A class which provides brackets, as well as separators
    and a function to facilitate formatting of the outer template.
    """

    #: Left bracket.
    left: Union["BaseText", str] = "["

    #: Right bracket.
    right: Union["BaseText", str] = "]"

    #: Separators used for outer template (i.e. in between references
    #: if multiple keys are referenced in a single citation).
    sep: Union["BaseText", str] = ", "

    #: Separator for outer template, if only two items.
    sep2: Optional[Union["BaseText", str]] = None

    #: Separator for outer template, for last item if three or more items.
    last_sep: Optional[Union["BaseText", str]] = None

    def outer(
        self, children: List["BaseText"], brackets=False, capfirst=False
    ) -> "Node":
        """Creates an outer template with separators,
        adding brackets if requested,
        and capitalizing the first word if requested.
        """
        return join[
            self.left if brackets else "",
            sentence(
                capfirst=capfirst,
                add_period=False,
                sep=self.sep,
                sep2=self.sep2,
                last_sep=self.last_sep,
            )[children],
            self.right if brackets else "",
        ]


@dataclass
class PersonStyle:
    """A class providing additional data and helper functions
    to facilitate formatting of person names.
    """

    #: Plugin name of the style used for formatting person names.
    style: str = "last"

    #: Plugin class instance used for formatting person names.
    #: Automatically initialised from :attr:`style`.
    style_plugin: "BaseNameStyle" = field(init=False)

    #: Whether or not to abbreviate first names.
    abbreviate: bool = True

    #: Separator between persons.
    sep: Union["BaseText", str] = ", "

    #: Separator between persons, if only two persons.
    sep2: Optional[Union["BaseText", str]] = " and "

    #: Separator between persons, for last person if three or more persons.
    last_sep: Optional[Union["BaseText", str]] = ", and "

    #: Abbreviation text if three or more persons.
    other: Optional[Union["BaseText", str]] = field(
        default_factory=lambda: Text(" ", Tag("em", "et al."))
    )

    def __post_init__(self):
        self.style_plugin = pybtex.plugin.find_plugin(
            "pybtex.style.names", name=self.style
        )()

    def names(self, role: str, full: bool) -> "Node":
        """Returns a template formatting the persons with correct separators
        and using the full person list if so requested.
        """
        return names(
            role=role,
            sep=self.sep,
            sep2=self.sep2,
            last_sep=self.last_sep,
            other=None if full else self.other,
        )

    def author_or_editor_or_title(self, full: bool) -> "Node":
        """Returns a template formatting the author, falling back on editor
        or title if author is not specified.
        """
        return author_or_editor_or_title(
            sep=self.sep,
            sep2=self.sep2,
            last_sep=self.last_sep,
            other=None if full else self.other,
        )


@dataclass
class GroupReferenceStyle(BaseReferenceStyle):
    """Composes a group of reference styles into a single consistent style."""

    #: List of style types.
    styles: List[BaseReferenceStyle] = field(default_factory=list)

    #: Dictionary from role names to styles.
    #: Automatically initialized from :attr:`styles`.
    role_style: Dict[str, BaseReferenceStyle] = field(default_factory=dict)

    def __post_init__(self):
        super().__post_init__()
        self.role_style.update(
            (role_name, style)
            for style in self.styles
            for role_name in style.role_names()
        )

    def role_names(self):
        return self.role_style.keys()

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        """Gets the outer template associated with *role_name*
        in one of the :attr:`styles`.
        """
        style = self.role_style[role_name]
        return style.outer(role_name, children)

    def inner(self, role_name: str) -> "Node":
        """Gets the inner template associated with *role_name*
        in one of the :attr:`styles`.
        """
        style = self.role_style[role_name]
        return style.inner(role_name)



"""Test back references."""

from test.common import html_citations

import pytest


@pytest.mark.sphinx("html", testroot="backrefs")
def test_backrefs(app, warning) -> None:
    app.build()
    output = (app.outdir / "index.html").read_text()
    match = html_citations(text=".*Test zero.*").search(output)
    assert match
    assert match.group("backref") is None
    assert match.group("backref1") is None
    assert match.group("backref2") is None
    assert match.group("backref3") is None
    match = html_citations(text=".*Test one.*").search(output)
    assert match
    assert match.group("backref") is not None
    assert match.group("backref1") is None
    assert match.group("backref2") is None
    assert match.group("backref3") is None
    match = html_citations(text=".*Test two.*").search(output)
    assert match
    assert match.group("backref") is None
    assert match.group("backref1") is not None
    assert match.group("backref2") is not None
    assert match.group("backref3") is None
    match = html_citations(text=".*Test three.*").search(output)
    assert match
    assert match.group("backref") is None
    assert match.group("backref1") is not None
    assert match.group("backref2") is not None
    assert match.group("backref3") is not None



import pytest

from sphinxcontrib.bibtex.plugin import find_plugin, register_plugin


def test_plugin_bad_group() -> None:
    with pytest.raises(ImportError, match="plugin group blablabla not found"):
        find_plugin("blablabla", "boo")
    with pytest.raises(ImportError, match="plugin group blablabla not found"):
        register_plugin("blablabla", "boo", type(None))


def test_plugin_register_not_forced() -> None:
    class Plugin:
        pass

    assert not register_plugin(
        "sphinxcontrib.bibtex.style.referencing", "label", Plugin
    )
    assert find_plugin("sphinxcontrib.bibtex.style.referencing", "label") is not Plugin


def test_plugin_register_forced() -> None:
    class PluginOld:
        pass

    class PluginNew:
        pass

    assert register_plugin(
        "sphinxcontrib.bibtex.style.referencing",
        "xxx_test_plugin_register_forced",
        PluginOld,
    )
    assert (
        find_plugin(
            "sphinxcontrib.bibtex.style.referencing", "xxx_test_plugin_register_forced"
        )
        is PluginOld
    )
    assert not register_plugin(
        "sphinxcontrib.bibtex.style.referencing",
        "xxx_test_plugin_register_forced",
        PluginNew,
    )
    assert (
        find_plugin(
            "sphinxcontrib.bibtex.style.referencing", "xxx_test_plugin_register_forced"
        )
        is PluginOld
    )
    assert register_plugin(
        "sphinxcontrib.bibtex.style.referencing",
        "xxx_test_plugin_register_forced",
        PluginNew,
        force=True,
    )
    assert (
        find_plugin(
            "sphinxcontrib.bibtex.style.referencing", "xxx_test_plugin_register_forced"
        )
        is PluginNew
    )



"""Some common helper functions for the test suite."""

import re
from typing import Optional

import docutils

RE_ID = r"[a-z][-?a-z0-9]*"
RE_NUM = r"\d+"
RE_LABEL = r"[^<]+"
RE_TEXT = r".*"
RE_DOCNAME = r"[^:]+"
RE_TITLE = r'[^"]*'


def html_citation_refs(refid=RE_ID, label=RE_LABEL, title: Optional[str] = RE_TITLE):
    title_pattern = rf' title="(?P<title>{title})"' if title is not None else ""
    return re.compile(
        r'<a class="reference internal"'
        r' href="(?P<refdoc>[^#]+)?#(?P<refid>{refid})"'
        r"{title_pattern}"
        r">"
        r"(?P<label>{label})"
        r"</a>".format(refid=refid, label=label, title_pattern=title_pattern)
    )


# match single citation with square brackets
# also gets the id of the citation itself (which will appear in backref)
def html_citation_refs_single(
    id_=RE_ID, refid=RE_ID, label=RE_LABEL, title: Optional[str] = RE_TITLE
):
    title_pattern = rf' title="{title}"' if title is not None else ""
    return re.compile(
        r'<span id="(?P<id_>{id_})">\['
        r'<a class="reference internal"'
        r' href="(?P<refdoc>[^#]+)?#(?P<refid>{refid})"'
        r"{title_pattern}"
        r">"
        r"(?P<label>{label})"
        r"</a>"
        r"]</span>".format(
            id_=id_, refid=refid, label=label, title_pattern=title_pattern
        )
    )


def html_docutils_citation_refs(refid=RE_ID, label=RE_LABEL, id_=RE_ID):
    return re.compile(
        r'<a class="reference internal" '
        r'href="(?P<refdoc>[^#]+)?#(?P<refid>{refid})" '
        r'id="(?P<id_>{id_})">'
        r"<span>\[(?P<label>{label})]</span>"
        r"</a>".format(refid=refid, label=label, id_=id_)
    )


def html_citations(id_=RE_ID, label=RE_LABEL, text=RE_TEXT):
    if docutils.__version_info__ < (0, 18):
        return re.compile(
            r'<dt class="label" id="(?P<id_>{id_})">'
            r'<span class="brackets">'
            r'(?:<a class="fn-backref" href="#(?P<backref>{backref_id})">)?'
            r"(?P<label>{label})"
            r"(?:</a>)?"
            r"</span>"
            r'(?:<span class="fn-backref">\('
            r'<a href="#(?P<backref1>{backref_id})">1</a>'
            r',<a href="#(?P<backref2>{backref_id}\w+)">2</a>'
            r'(,<a href="#(?P<backref3>{backref_id}\w+)">3</a>)?'
            r'(,<a href="#\w+">\d+</a>)*'  # no named group for extra backrefs
            r"\)</span>)?"
            r"</dt>\n"
            r"<dd><p>(?P<text>{text})</p>\n</dd>".format(
                id_=id_, label=label, text=text, backref_id=RE_ID
            )
        )
    else:
        return re.compile(
            r'<div class="citation" id="(?P<id_>{id_})"'
            r' role="doc-biblioentry">\s*'
            r'<span class="label">'
            r'<span class="fn-bracket">\[</span>'
            r'(?:<a role="doc-backlink" href="#(?P<backref>{backref_id})">)?'
            r"(?P<label>{label})"
            r"(?:</a>)?"
            r'<span class="fn-bracket">]</span>'
            r"</span>\s*"
            r'(?:<span class="backrefs">\('
            r'<a {back_role} href="#(?P<backref1>{backref_id})">1</a>'
            r',<a {back_role} href="#(?P<backref2>{backref_id}\w+)">2</a>'
            r'(,<a {back_role} href="#(?P<backref3>{backref_id}\w+)">3</a>)?'
            r'(,<a {back_role} href="#\w+">\d+</a>)*'
            r"\)</span>\s*)?"
            r"<p>(?P<text>{text})</p>\s*"
            r"</div>".format(
                back_role='role="doc-backlink"',
                id_=id_,
                label=label,
                text=text,
                backref_id=RE_ID,
            )
        )


def html_footnote_refs(refid=RE_ID):
    if docutils.__version_info__ < (0, 18):
        return re.compile(
            r'<a class="footnote-reference brackets"'
            r' href="#(?P<refid>{refid})" id="(?P<id_>{id_})">'
            r"(?P<label>{label})"
            r"</a>".format(refid=refid, id_=RE_ID, label=RE_NUM)
        )
    else:
        return re.compile(
            r'<a class="footnote-reference brackets" '
            r'href="#(?P<refid>{refid})" id="(?P<id_>{id_})" '
            r'role="doc-noteref">'
            r'<span class="fn-bracket">\[</span>'
            r"(?P<label>{label})"
            r'<span class="fn-bracket">]</span>'
            r"</a>".format(refid=refid, id_=RE_ID, label=RE_NUM)
        )


def html_footnotes(id_=RE_ID, text=RE_TEXT):
    if docutils.__version_info__ < (0, 18):
        return re.compile(
            r'<dt class="label" id="(?P<id_>{id_})">'
            r'<span class="brackets">'
            r'(?:<a class="fn-backref" href="#(?P<backref>{backref_id})">)?'
            r"(?P<label>{label})"
            r"(?:</a>)?"
            r"</span>"
            r'(?:<span class="fn-backref">\('
            r'<a href="#(?P<backref1>{backref_id})">1</a>'
            r',<a href="#(?P<backref2>{backref_id}\w+)">2</a>'
            r'(,<a href="#(?P<backref3>{backref_id}\w+)">3</a>)?'
            r'(,<a href="#\w+">\d+</a>)*'
            r"\)</span>)?"
            r"</dt>\n"
            r"<dd><p>(?P<text>{text})</p>\n</dd>".format(
                id_=id_, backref_id=RE_ID, label=RE_NUM, text=text
            )
        )
    else:
        foot_role = (
            'role="note"'
            if docutils.__version_info__ < (0, 20)
            else 'role="doc-footnote"'
        )
        return re.compile(
            r'<aside class="footnote brackets" id="(?P<id_>{id_})"'
            r" {foot_role}>\s*"
            r'<span class="label">'
            r'<span class="fn-bracket">\[</span>'
            r'(?:<a {back_role} href="#(?P<backref>{backref_id})">)?'
            r"{label}"
            r"(?:</a>)?"
            r'<span class="fn-bracket">]</span>'
            r"</span>\s*"
            r'(?:<span class="backrefs">\('
            r'<a {back_role} href="#(?P<backref1>{backref_id})">1</a>'
            r',<a {back_role} href="#(?P<backref2>{backref_id}\w+)">2</a>'
            r'(,<a {back_role} href="#(?P<backref3>{backref_id}\w+)">3</a>)?'
            r'(,<a {back_role} href="#\w+">\d+</a>)*'
            r"\)</span>\s*)?"
            r"<p>(?P<text>{text})</p>\s*"
            r"</aside>".format(
                foot_role=foot_role,
                back_role='role="doc-backlink"',
                id_=id_,
                label=RE_NUM,
                text=text,
                backref_id=RE_ID,
            )
        )


def latex_citations(docname=RE_DOCNAME, id_=RE_ID, label=RE_LABEL, text=RE_TEXT):
    return re.compile(
        r"\\bibitem\[(?P<label>{label})]"
        r"{{(?P<docname>{docname}):(?P<id_>{id_})}}\n"
        r"\\sphinxAtStartPar\n"
        r"(?P<text>{text})\n".format(docname=docname, label=label, id_=id_, text=text)
    )


def latex_citation_refs(docname=RE_DOCNAME, refid=RE_ID):
    return re.compile(
        rf"\\hyperlink{{cite[.](?P<docname>{docname}):(?P<refid>{refid})}}"
    )



"""
    .. autofunction:: setup
"""

from typing import Any, Dict

from sphinx.application import Sphinx

from .directives import BibliographyDirective
from .domain import BibtexDomain
from .foot_directives import FootBibliographyDirective
from .foot_domain import BibtexFootDomain
from .foot_roles import FootCiteRole
from .nodes import bibliography, depart_raw_latex, raw_latex, visit_raw_latex
from .roles import CiteRole
from .transforms import BibliographyTransform


def setup(app: Sphinx) -> Dict[str, Any]:
    """Set up the bibtex extension:

    * register config values
    * register directives
    * register nodes
    * register roles
    * register transforms
    * connect events to functions
    """
    app.add_config_value("bibtex_default_style", "alpha", "html")
    app.add_config_value("bibtex_tooltips", True, "html")
    app.add_config_value("bibtex_tooltips_style", "", "html")
    app.add_config_value("bibtex_bibfiles", None, "html")
    app.add_config_value("bibtex_encoding", "utf-8-sig", "html")
    app.add_config_value("bibtex_bibliography_header", "", "html")
    app.add_config_value("bibtex_footbibliography_header", "", "html")
    app.add_config_value("bibtex_reference_style", "label", "env")
    app.add_config_value("bibtex_foot_reference_style", "foot", "env")
    app.add_config_value("bibtex_cite_id", "", "html")
    app.add_config_value("bibtex_footcite_id", "", "html")
    app.add_config_value("bibtex_bibliography_id", "", "html")
    app.add_config_value("bibtex_footbibliography_id", "", "html")
    app.add_domain(BibtexDomain)
    app.add_directive("bibliography", BibliographyDirective)
    app.add_role("cite", CiteRole())
    app.add_node(bibliography, override=True)
    app.add_node(raw_latex, latex=(visit_raw_latex, depart_raw_latex), override=True)
    app.add_post_transform(BibliographyTransform)
    app.add_domain(BibtexFootDomain)
    app.add_directive("footbibliography", FootBibliographyDirective)
    app.add_role("footcite", FootCiteRole())

    return {
        "version": "2.6.2",
        "env_version": 9,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
    test_sphinx
    ~~~~~~~~~~~

    General Sphinx test and check output.
"""

import pytest


@pytest.mark.sphinx("html", testroot="sphinx")
def test_sphinx(app, warning) -> None:
    app.build()
    assert not warning.getvalue()



import re

import pytest


@pytest.mark.rinohtype
@pytest.mark.sphinx("rinoh", testroot="citation_rinoh")
def test_citation_rinoh(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "book.stylelog").read_text()
    assert len(re.findall(r"NoteMarkerByID\('tes', style='citation'\)", output)) == 1
    assert len(re.findall(r"DirectReference\('tes'\)", output)) == 1


@pytest.mark.rinohtype
@pytest.mark.sphinx("rinoh", testroot="citation_rinoh_multidoc")
def test_citation_rinoh_multidoc(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "book.stylelog").read_text()
    assert len(re.findall(r"NoteMarkerByID\('tes', style='citation'\)", output)) == 2
    assert len(re.findall(r"DirectReference\('tes'\)", output)) == 1



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, List, Union

from pybtex.style.template import tag

from sphinxcontrib.bibtex.style.template import (
    entry_label,
    join,
    join2,
    post_text,
    pre_text,
    reference,
)

from . import BaseReferenceStyle, BracketStyle, PersonStyle

if TYPE_CHECKING:
    from pybtex.richtext import BaseText
    from pybtex.style.template import Node


@dataclass
class BasicSuperParentheticalReferenceStyle(BaseReferenceStyle):
    """Reference by label superscript if parenthetical,
    and by author and label superscript if textual.
    """

    #: Bracket style. Left and right brackets are empty by default.
    bracket: BracketStyle = field(
        default_factory=lambda: BracketStyle(left="", right="", sep=",")
    )

    #: Person style.
    person: PersonStyle = field(default_factory=PersonStyle)

    #: Separator between pre-text and citation.
    pre_text_sep: Union["BaseText", str] = " "

    #: Separator between citation and post-text.
    post_text_sep: Union["BaseText", str] = ", "

    def role_names(self) -> Iterable[str]:
        return [
            f"{alt}p{full_author}" for alt in ["", "al"] for full_author in ["", "s"]
        ]

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        if "al" not in role_name:
            return tag("sup")[
                self.bracket.outer(children, brackets=True, capfirst=False)
            ]
        else:
            return self.bracket.outer(children, brackets=False, capfirst=False)

    def inner(self, role_name: str) -> "Node":
        return join2(sep1=self.pre_text_sep, sep2=self.post_text_sep)[
            pre_text,
            reference[entry_label],
            post_text,
        ]


@dataclass
class BasicSuperTextualReferenceStyle(BaseReferenceStyle):
    """Reference by label superscript if parenthetical,
    and by author and label superscript if textual.
    """

    #: Bracket style. Left and right brackets are empty by default.
    bracket: BracketStyle = field(
        default_factory=lambda: BracketStyle(left="", right="", sep=", ")
    )

    #: Person style.
    person: PersonStyle = field(default_factory=PersonStyle)

    #: Separator between text and reference.
    text_reference_sep: Union["BaseText", str] = ""

    #: Separator between pre-text and citation.
    pre_text_sep: Union["BaseText", str] = " "

    #: Separator between citation and post-text.
    post_text_sep: Union["BaseText", str] = ", "

    def role_names(self) -> Iterable[str]:
        return [
            f"{capfirst}t{full_author}"
            for capfirst in ["", "c"]
            for full_author in ["", "s"]
        ]

    def outer(self, role_name: str, children: List["BaseText"]) -> "Node":
        return self.bracket.outer(children, brackets=False, capfirst="c" in role_name)

    def inner(self, role_name: str) -> "Node":
        return join(sep=self.text_reference_sep)[
            self.person.author_or_editor_or_title(full="s" in role_name),
            tag("sup")[
                join[
                    self.bracket.left,
                    join2(sep1=self.pre_text_sep, sep2=self.post_text_sep)[
                        pre_text,
                        reference[entry_label],
                        post_text,
                    ],
                    self.bracket.right,
                ]
            ],
        ]



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Union

from . import BracketStyle, GroupReferenceStyle, PersonStyle
from .basic_foot import (
    BasicFootParentheticalReferenceStyle,
    BasicFootTextualReferenceStyle,
)

if TYPE_CHECKING:
    from pybtex.richtext import BaseText


@dataclass
class FootReferenceStyle(GroupReferenceStyle):
    """Textual or parenthetical reference using footnotes."""

    #: Bracket style for textual citations (:cite:t: and variations).
    bracket_textual: BracketStyle = field(default_factory=BracketStyle)

    #: Person style (applies to all relevant citation commands).
    person: PersonStyle = field(default_factory=PersonStyle)

    #: Separator between text and reference for textual citations.
    text_reference_sep: Union["BaseText", str] = ""

    def __post_init__(self):
        self.styles.extend(
            [
                BasicFootParentheticalReferenceStyle(),
                BasicFootTextualReferenceStyle(
                    bracket=self.bracket_textual,
                    person=self.person,
                    text_reference_sep=self.text_reference_sep,
                ),
            ]
        )
        super().__post_init__()



extensions = ["sphinxcontrib.bibtex"]
exclude_patterns = ["_build"]
bibtex_bibfiles = ["test.bib"]
bibtex_bibliography_header = ".. rubric:: Regular Citations"
bibtex_footbibliography_header = ".. rubric:: Footnote Citations"



import sys

if sys.version_info >= (3, 10):
    from importlib.metadata import EntryPoint, entry_points
else:
    from importlib_metadata import entry_points, EntryPoint

from typing import Any, Dict, List, Type

_runtime_plugins: Dict[str, Dict[str, Type]] = {
    "sphinxcontrib.bibtex.style.referencing": {}
}


# wrapper to work around missing type annotations for entry_points function
def _entry_points(group: str, name: str) -> List[EntryPoint]:
    return entry_points(group=group, name=name)  # type: ignore


def find_plugin(group: str, name: str) -> Type[Any]:
    """Load a sphinxcontrib-bibtex plugin, either from the runtime store,
    or from the entry points.
    """
    global _runtime_plugins
    if group not in _runtime_plugins:
        raise ImportError(f"plugin group {group} not found")
    try:
        return _runtime_plugins[group][name]
    except KeyError:
        for entry_point in _entry_points(group=group, name=name):
            return entry_point.load()
    raise ImportError(f"plugin {group}.{name} not found")


def register_plugin(
    group: str, name: str, klass: Type[Any], force: bool = False
) -> bool:
    """Register a sphinxcontrib-bibtex plugin into the runtime store."""
    global _runtime_plugins
    if group not in _runtime_plugins:
        raise ImportError(f"plugin group {group} not found")
    eps: List[Any]
    try:
        eps = [_runtime_plugins[group][name]]
    except KeyError:
        eps = _entry_points(group=group, name=name)
    if not eps or force:
        _runtime_plugins[group][name] = klass
        return True
    else:
        return False



# patched version of Weston Nielson's sphinx-natbib

# changes:
#   * fix some typos and flake warnings, add a few type annotations
#   * remove unused latex_to_nodes and SUBSUP_RE
#   * remove latex special casing to fix build error
#   * simplify super calls
#   * fix refdoc bug
#   * use XRefRole.result_nodes instead of XRefRole.__call__
#   * use dictionary as ordered set (assuming Python 3.6+)

from typing import List, Tuple, cast

import docutils.parsers.rst.directives as directives
import pybtex.backends.plaintext
import pybtex.style.names.lastfirst
import pybtex.style.names.plain
from docutils import nodes
from docutils.parsers.rst import Directive
from pybtex.database.input import bibtex
from sphinx import addnodes
from sphinx.domains import Domain, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.locale import _
from sphinx.roles import XRefRole
from sphinx.util import logging

logger = logging.getLogger(__name__)

DEFAULT_CONF = {
    "file": "",
    "brackets": "()",
    "separator": ";",
    "style": "authoryear",  # 'numbers', 'super'
    "sort": False,
    "sort_compress": False,
}

ROLES = [
    "p",
    "ps",
    "alp",
    "alps",
    "t",
    "ts",
    "alt",
    "alts",
    "author",
    "authors",
    "year",
    "yearpar",
    "text",
    "title",
]


def parse_keys(rawtext) -> Tuple[List[str], str, str]:
    # Get the keys and any pre- and post-citation text
    # Spaces nor commas are allowed in cite keys, so we split on commas
    # first.  This will give us a list of keys, however, the last item may have
    # pre- and post-citation text (in brackets "[pre][post]")
    #
    # TODO: This isn't the best implementation and this should also
    #       handle errors
    pre = ""
    post = ""
    keys = []
    for k in rawtext.split(","):
        k = k.strip()  # Remove leading and trailing whitespace
        k = k.split(" ", 1)  # Split on the first space in the key, if any
        if len(k) > 1:
            # We have some extra text here
            k, text = k
            bo, bc = 0, 0
            for c in text.strip():
                if c == "[" and bo == bc:
                    bo += 1
                elif c == "]" and bo == bc + 1:
                    bc += 1
                else:
                    if bc == 0:
                        pre += c
                    else:
                        post += c
            if bc == bo and bc == 1:
                post = pre
                pre = ""
        else:
            k = k[0]
        keys.append(k)
    return keys, pre, post


class Citations:
    def __init__(self, env):
        self.conf = DEFAULT_CONF.copy()
        self.conf.update(env.app.config.natbib)

        self.file_name = None
        self.parser = None
        self.data = None
        self.ref_map = {}
        self.file_name = env.relfn2path(self.conf["file"], env.app.config.master_doc)[1]
        self.parser = bibtex.Parser()
        self.data = self.parser.parse_file(self.file_name)

    def get(self, key):
        return self.data.entries.get(key)


class CitationTransform:
    """
    This class is meant to be applied to a ``docutils.nodes.pending`` node when
    a ``cite`` role is encountered.  Later (during the resolve_xref stage) this
    class can be used to generate the proper citation reference nodes for
    insertion into the actual document.
    """

    def __init__(self, refs, pre, post, typ, global_keys, config):
        self.refs = refs
        self.pre = pre
        self.post = post
        self.config = config
        self.typ = typ
        self.global_keys = global_keys

    def __repr__(self):
        return "<%s>" % self.__str__()

    def __str__(self):
        return ",".join([r.key for r in self.refs])

    def get_ref_num(self, key):
        return list(self.global_keys.keys()).index(key) + 1

    def get_author(self, authors, all_authors=False):
        if len(authors) == 0:
            author = ""
        elif len(authors) > 2 and not all_authors:
            author = "%s et al." % authors[0].last_names[0]
        elif len(authors) == 1:
            author = authors[0].last_names[0]
        else:
            author = "%s and %s" % (
                ", ".join([a.last_names[0] for a in authors[:-1]]),
                authors[-1].last_names[0],
            )
        author = author.replace("{", "")
        author = author.replace("}", "")
        return author

    # TODO refactor to reduce complexity, for now ignore C901
    def cite(self, cmd, refuri, global_keys):  # noqa: C901
        """
        Return a docutils Node consisting of properly formatted citations
        children nodes.
        """
        self.global_keys = global_keys
        bo, bc = self.config["brackets"]
        sep = "%s " % self.config["separator"]
        style = self.config["style"]
        all_auths = cmd.endswith("s")

        if (cmd.startswith("p") or cmd == "yearpar") and style != "super":
            node = nodes.inline(bo, bo, classes=["citation"])
        else:
            node = nodes.inline("", "", classes=["citation"])

        if self.pre:
            pre = "%s " % self.pre
            node += nodes.inline(pre, pre, classes=["pre"])

        for i, ref in enumerate(self.refs):
            authors = ref.persons.get("author", [])
            author_text = self.get_author(authors, all_auths)
            lrefuri = refuri + "#citation-" + nodes.make_id(ref.key)

            if 0 < i < len(self.refs):
                if style == "authoryear":
                    node += nodes.inline(sep, sep)
                else:
                    if style == "super":
                        node += nodes.superscript(", ", ", ")
                    else:
                        node += nodes.inline(", ", ", ")

            if cmd == "title":
                title = ref.fields.get("title")
                if title is None:
                    title = ref.fields.get("key", "")
                author_text = title

            if (
                style == "authoryear" and (cmd.startswith("p") or cmd.startswith("alp"))
            ) or (
                cmd.startswith("t") or cmd.startswith("alt") or cmd.startswith("author")
            ):
                node += nodes.reference(
                    author_text, author_text, internal=True, refuri=lrefuri
                )

                if cmd.startswith("p") or cmd.startswith("alp"):
                    node += nodes.inline(", ", ", ")
                else:
                    node += nodes.inline(" ", " ")

            # Add in either the year or the citation number
            if cmd == "title":
                pass
            elif cmd.startswith("author"):
                pass
            else:
                if style != "authoryear":
                    num = self.get_ref_num(ref.key)
                else:
                    num = ref.fields.get("year")

                refnode = nodes.reference(
                    str(num), str(num), internal=True, refuri=lrefuri
                )

                if cmd.startswith("t") and style != "super":
                    node += nodes.inline(bo, bo)

                if style == "super":
                    node += nodes.superscript("", "", refnode)
                else:
                    node += refnode

                if cmd.startswith("t") and style != "super":
                    node += nodes.inline(bc, bc)

        if self.post:
            post = ", %s" % self.post
            node += nodes.inline(post, post, classes=["post"])

        if (cmd.startswith("p") or cmd == "yearpar") and style != "super":
            node += nodes.inline(bc, bc, classes=["citation"])

        return node


def sort_references(refs, citations):
    def sortkey(key):
        # sort by author last names, but if no author, sort by title
        citation = citations.get(key)
        authorsort = "".join(map(str, citation.persons.get("author", "")))
        if len(authorsort) > 0:
            authorsort = authorsort.replace("{", "")
            authorsort = authorsort.replace("}", "")
            return authorsort.upper()
        titlesort = citation.fields.get("title", "")
        titlesort = titlesort.replace("{", "")
        titlesort = titlesort.replace("}", "")
        return titlesort.upper()

    sortedrefs = sorted(refs, key=sortkey)
    return {ref: None for ref in sortedrefs}


class CitationXRefRole(XRefRole):
    def result_nodes(
        self,
        document: nodes.document,
        env: BuildEnvironment,
        node: nodes.Element,
        is_ref: bool,
    ) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """
        When a ``cite`` role is encountered, we replace it with a
        ``docutils.nodes.pending`` node that uses a ``CitationTransform`` for
        generating the proper citation reference representation during the
        resolve_xref phase.
        """
        domain = cast(CitationDomain, env.get_domain("cite"))

        # Get the config at this point in the document
        config = {}
        for opt in ["style", "brackets", "separator", "sort", "sort_compress"]:
            config[opt] = env.temp_data.get(
                "cite_%s" % opt,
                env.domaindata["cite"]["conf"].get(opt, DEFAULT_CONF[opt]),
            )

        if self.name == "cite:text":
            # A ``text`` citation is unique because it doesn't reference a
            # cite-key
            keys: List[str] = []
            pre, post = self.text, ""
        else:
            keys, pre, post = parse_keys(self.text)
            for key in keys:
                if domain.citations.get(key) is None:
                    logger.warning(
                        "cite-key `%s` not found in bibtex file" % key,
                        location=(env.docname, self.lineno),
                    )
                    continue
                env.domaindata["cite"]["keys"][key] = None
                env.domaindata["cite"]["keys"] = sort_references(
                    env.domaindata["cite"]["keys"], domain.citations
                )

        data = {
            "keys": keys,
            "pre": pre,
            "post": post,
            "typ": self.name,
            "global_keys": env.domaindata["cite"]["keys"],
            "config": config,
        }

        # TODO bad design here, pending is meant to get the Transform object
        node += nodes.pending(CitationTransform, data)  # type: ignore
        return [node], []


class CitationConfDirective(Directive):
    """
    Allows the user to change the citation style on a per-page or
    per-block basis.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 1
    option_spec = {
        "brackets": directives.unchanged,
        "separator": directives.unchanged,
        "style": directives.unchanged,
        "sort": directives.flag,
        "sort_compress": directives.flag,
    }

    def run(self):
        env = self.state.document.settings.env

        # TODO: verify options
        if self.arguments:
            env.temp_data["cite_style"] = self.arguments[0]
        else:
            env.temp_data["cite_style"] = self.options.get(
                "style", DEFAULT_CONF["style"]
            )

        try:
            self.options.pop("style")
        except KeyError:
            pass

        for k, v in self.options.items():
            env.temp_data["cite_%s" % k] = v

        return []


class CitationReferencesDirective(Directive):
    """
    Generates the actual reference list.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 0

    # TODO: Implement support for multiple bib files
    option_spec = {
        "path": directives.unchanged,
    }

    # TODO refactor to reduce complexity, for now ignore C901
    def get_reference_node(self, ref):  # noqa: C901
        node = nodes.inline("", "", classes=[ref.type, "reference"])

        namestyler = pybtex.style.names.plain.NameStyle()
        plaintext = pybtex.backends.plaintext.Backend()

        # Authors
        authors = ref.persons.get("author", [])
        for i, author in enumerate(authors):
            authortext = namestyler.format(author, abbr=True).format().render(plaintext)
            authortext = authortext.replace("{", "")
            authortext = authortext.replace("}", "")
            text = authortext

            text = text.strip()
            auth_node = nodes.inline(text, text)
            auth_node["classes"].append("author")
            node += auth_node

            if i + 1 < len(authors):
                node += nodes.inline(", ", ", ")
            else:
                ending = "%s  " % ("" if text.endswith(".") else ".")
                node += nodes.inline(ending, ending)

        # Title
        title = ref.fields.get("title")
        if title is None:
            title = ref.fields.get("key")
        if title:
            title = title.replace("{", "")
            title = title.replace("}", "")
            node += nodes.inline(title, title, classes=["title"])
            node += nodes.inline(".  ", ".  ")

        # @phdthesis
        if ref.type == "phdthesis":
            school = ref.fields.get("school")
            text = "PhD Thesis, %s, " % school
            node += nodes.inline(text, text)

        # Publication
        pub = ref.fields.get("journal")
        if not pub:
            pub = ref.fields.get("booktitle")
        if pub:
            pub = pub.replace("{", "")
            pub = pub.replace("}", "")
            node += nodes.emphasis(pub, pub, classes=["publication"])
            node += nodes.inline(", ", ", ")

        vol = ref.fields.get("volume")
        pages = ref.fields.get("pages")
        year = ref.fields.get("year")

        if pub is None:
            howpub = ref.fields.get("howpublished")
            if howpub is not None and howpub.startswith("\\url{"):
                url = howpub[5:-1]
                refnode = nodes.reference("", "", internal=False, refuri=url)
                refnode += nodes.Text(url)
                node += refnode
                if vol or pages or year:
                    node += nodes.inline(", ", ", ")

        if vol:
            node += nodes.inline(vol, vol, classes=["volume"])
            node += nodes.inline(":", ":")

        if pages:
            node += nodes.inline(pages, pages, classes=["pages"])
            node += nodes.inline(", ", ", ")

        if year:
            node += nodes.inline(year, year, classes=["year"])
            node += nodes.inline(".", ".")

        return node

    def run(self):
        """
        Generate the definition list that displays the actual references.
        """
        env = self.state.document.settings.env
        keys = env.domaindata["cite"]["keys"]
        env.domaindata["cite"]["refdoc"] = env.docname

        citations = env.domains["cite"].citations

        tbody = nodes.tbody("")
        for i, key in enumerate(keys):
            row = nodes.row("")
            nid = "citation-%s" % nodes.make_id(key)
            row["classes"].append("footnote")
            row["ids"].append(nid)
            row["names"].append(nid)

            numcol = nodes.entry("", nodes.paragraph("", "[%d]" % (i + 1)))
            definition = self.get_reference_node(citations.get(key))
            refcol = nodes.entry("", nodes.paragraph("", "", definition))
            row.extend([numcol, refcol])

            tbody.append(row)

        table_spec_node = addnodes.tabular_col_spec()
        table_spec_node["spec"] = "cl"

        node = nodes.table(
            "",
            table_spec_node,
            nodes.tgroup(
                "",
                nodes.colspec(colwidth=10, classes=["label"]),
                nodes.colspec(colwidth=90),
                tbody,
            ),
        )

        return [node]


class CitationDomain(Domain):
    name = "cite"
    label = "citation"

    object_types = {
        "citation": ObjType(_("citation"), *ROLES, searchprio=-1),
    }

    directives = {"conf": CitationConfDirective, "refs": CitationReferencesDirective}
    roles = dict([(r, CitationXRefRole()) for r in ROLES])

    initial_data = {
        "keys": {},  # cite-keys in order of reference using dict as sorted set
        "conf": DEFAULT_CONF,
        "refdoc": None,
    }

    def __init__(self, env):
        super().__init__(env)

        # Update conf
        self.data["conf"].update(env.app.config.natbib)

        # TODO: warn if citations can't parse bibtex file
        self.citations = Citations(env)

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        refdoc = self.data["refdoc"]
        if not refdoc:
            logger.warning(
                "no `refs` directive found; citations will have dead links",
                location=node,
            )
            refuri = ""
        else:
            refuri = builder.get_relative_uri(fromdocname, refdoc)

        for nd in node.children:
            if isinstance(nd, nodes.pending):
                nd.details["refs"] = []

                for key in nd.details.pop("keys"):
                    ref = self.citations.get(key)
                    if ref is None:
                        continue
                    nd.details["refs"].append(ref)

                transform = nd.transform(**nd.details)
                node = transform.cite(typ, refuri, global_keys=self.data["keys"])

        return node


def setup(app):
    app.add_config_value("natbib", DEFAULT_CONF, "env")
    app.add_domain(CitationDomain)



import re
from test.common import html_citations

import pytest


@pytest.mark.sphinx("html", testroot="list_citation")
def test_list_citation(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    assert html_citations(label="1", text=".*Akkerdju.*").search(output)
    assert html_citations(label="2", text=".*Bro.*").search(output)
    assert html_citations(label="3", text=".*Chap.*").search(output)
    assert html_citations(label="4", text=".*Dude.*").search(output)


@pytest.mark.sphinx("html", testroot="list_bullet")
def test_list_bullet(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    assert re.search(
        "<ul"
        ".*<li.*Akkerdju.*</li>"
        ".*<li.*Bro.*</li>"
        ".*<li.*Chap.*</li>"
        ".*<li.*Dude.*</li>"
        ".*</ul>",
        output,
        re.DOTALL,
    )


@pytest.mark.sphinx("html", testroot="list_enumerated")
def test_list_enumerated(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.html").read_text()
    assert re.search(
        '<ol .*start="1".*>'
        ".*<li.*Akkerdju.*</li>"
        ".*<li.*Bro.*</li>"
        ".*<li.*Chap.*</li>"
        ".*<li.*Dude.*</li>"
        ".*</ol>"
        '.*<ol .*start="5".*>'
        ".*<li.*Eminence.*</li>"
        ".*<li.*Frater.*</li>"
        ".*<li.*Giggles.*</li>"
        ".*<li.*Handy.*</li>"
        ".*</ol>"
        '.*<ol .*start="23".*>'
        ".*<li.*Iedereen.*</li>"
        ".*<li.*Joke.*</li>"
        ".*<li.*Klopgeest.*</li>"
        ".*<li.*Laterfanter.*</li>"
        ".*</ol>",
        output,
        re.DOTALL,
    )


@pytest.mark.sphinx("html", testroot="list_invalid")
def test_list_invalid(app, warning) -> None:
    app.build()
    assert re.search(
        "unknown bibliography list type 'thisisintentionallyinvalid'",
        warning.getvalue(),
    )



from abc import ABC
from typing import Generic, TypeVar

from pybtex.richtext import BaseMultipartText, BaseText

ReferenceInfo = TypeVar("ReferenceInfo")
"""Generic type parameter for types that store reference information.
To be implemented by clients. See for instance
:class:`~sphinxcontrib.bibtex.domain.SphinxReferenceInfo`.
"""


class BaseReferenceText(BaseMultipartText, Generic[ReferenceInfo], ABC):
    """Generic rich text element for citation references.
    Instances store some extra reference info that can be used when formatting.
    This base class renders its children without further formatting.
    Implementations must create a derivation from this class which
    overrides the *render* method to create the desired output.
    See for instance
    :class:`~sphinxcontrib.bibtex.domain.SphinxReferenceText`.
    """

    def __init__(self, info: ReferenceInfo, *parts: "BaseText"):
        self.info = (info,)
        super().__init__(*parts)



from dataclasses import dataclass, field

import sphinxcontrib.bibtex.plugin
from sphinxcontrib.bibtex.style.referencing import BracketStyle
from sphinxcontrib.bibtex.style.referencing.author_year import AuthorYearReferenceStyle


def bracket_style() -> BracketStyle:
    return BracketStyle(
        left="(",
        right=")",
    )


@dataclass
class MyReferenceStyle(AuthorYearReferenceStyle):
    bracket_parenthetical: BracketStyle = field(default_factory=bracket_style)
    bracket_textual: BracketStyle = field(default_factory=bracket_style)
    bracket_author: BracketStyle = field(default_factory=bracket_style)
    bracket_label: BracketStyle = field(default_factory=bracket_style)
    bracket_year: BracketStyle = field(default_factory=bracket_style)


sphinxcontrib.bibtex.plugin.register_plugin(
    "sphinxcontrib.bibtex.style.referencing", "author_year_round", MyReferenceStyle
)


extensions = ["sphinxcontrib.bibtex"]
exclude_patterns = ["_build"]
bibtex_bibfiles = ["refs.bib"]
bibtex_reference_style = "author_year_round"



"""Docstring for some_module. :footcite:`testmodule`"""


def func(funcarg):
    """Docstring for function func.

    Long description goes here. See :footcite:`testfunc`.

    :param funcarg: Docstring for parameter. :footcite:`testfuncarg`
    """


a = 1
"""Docstring for variable a.

Long description goes here. See :footcite:`testdata`.
"""


class Foo:
    """Docstring for class Foo.

    Long description goes here. See :footcite:`testclass`.
    """

    b = 2
    """Docstring for class attribute b.

    Long description goes here. See :footcite:`testclassattr`.
    """

    def __init__(self, initarg):
        """Docstring for constructor.

        Long description goes here. See :footcite:`testinit`.

        :param initarg: Docstring for parameter. :footcite:`testinitarg`
        """

        self.c = 3
        """Docstring for instance attribute c.

        Long description goes here. See :footcite:`testinstanceattr`
        """

    def method(self, methodarg):
        """Docstring for method.

        Long description goes here. See  :footcite:`testmethod`.

        :param methodarg: Docstring for parameter. :footcite:`testmethodarg`
        """



import pytest

from sphinxcontrib.bibtex.citation_target import parse_citation_targets


@pytest.mark.parametrize(
    "target,expected",
    [
        ("abc", [("abc", "", "")]),
        ("{xyz}abc", [("abc", "xyz", "")]),
        ("abc{xyz}", [("abc", "", "xyz")]),
        ("{xyz}abc{123}", [("abc", "xyz", "123")]),
        ("  {xyz} abc  {123}   ", [("abc", "xyz", "123")]),
        ("abc,def", [("abc", "", ""), ("def", "", "")]),
        ("  abc ,   def  ", [("abc", "", ""), ("def", "", "")]),
        (
            "  {xyz }abc  {123} ,  {hi hi} def {oops, blah } ",
            [("abc", "xyz ", "123"), ("def", "hi hi", "oops, blah ")],
        ),
    ],
)
def test_citation_target(target, expected) -> None:
    assert list(parse_citation_targets(target)) == expected


@pytest.mark.parametrize(
    "target",
    [
        "}",
        "{}}",
        "abc}",
        "abc{}}",
        "",
        ",",
        ",a",
        "a,",
        "{}a",
        "a{}",
        "{a}b{}",
        "{}a{b}",
    ],
)
def test_citation_target_invalid(target) -> None:
    with pytest.raises(ValueError, match="malformed citation target"):
        list(parse_citation_targets(target))



import sphinx.builders.latex.transforms

extensions = ["sphinxcontrib.bibtex"]
bibtex_bibfiles = ["sources.bib"]
latex_documents = [
    ("index", "test.tex", "Test", "Mr. Test", "manual"),
]


class DummyTransform(sphinx.builders.latex.transforms.BibliographyTransform):
    def run(self, **kwargs):
        pass


sphinx.builders.latex.transforms.BibliographyTransform = DummyTransform



import re
from dataclasses import dataclass, field

import pytest

import sphinxcontrib.bibtex.plugin
from sphinxcontrib.bibtex.style.referencing import BracketStyle, PersonStyle
from sphinxcontrib.bibtex.style.referencing.foot import FootReferenceStyle


@pytest.mark.sphinx("text", testroot="footcite_roles")
def test_footcite_roles(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    tests = [
        ("p", " [1] "),
        ("ps", " [1] "),
        ("t", " de Du *et al.*[1] "),
        ("ts", " de Du, Em, and Fa[1] "),
        ("ct", " De Du *et al.*[1] "),
        ("cts", " De Du, Em, and Fa[1] "),
        ("p", " [2][3] "),
        ("ps", " [2][3] "),
        ("t", " al Ap[2], Be and Ci[3] "),
        ("ts", " al Ap[2], Be and Ci[3] "),
        ("ct", " Al Ap[2], Be and Ci[3] "),
        ("cts", " Al Ap[2], Be and Ci[3] "),
        ("p", " [4][5][6] "),
        ("ps", " [4][5][6] "),
        ("t", " Ge[4], Hu[5], Ix[6] "),
        ("ts", " Ge[4], Hu[5], Ix[6] "),
        ("ct", " Ge[4], Hu[5], Ix[6] "),
        ("cts", " Ge[4], Hu[5], Ix[6] "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":footcite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None


def my_bracket() -> BracketStyle:
    return BracketStyle(
        sep="; ",
        sep2="; ",
        last_sep="; ",
    )


def my_person() -> PersonStyle:
    return PersonStyle(
        style="last",
        abbreviate=False,
        sep=" & ",
        sep2=None,
        last_sep=None,
        other=" et al",
    )


@dataclass
class CustomReferenceStyle(FootReferenceStyle):
    bracket_textual: BracketStyle = field(default_factory=my_bracket)
    person: PersonStyle = field(default_factory=my_person)


sphinxcontrib.bibtex.plugin.register_plugin(
    "sphinxcontrib.bibtex.style.referencing",
    "xxx_foot_custom_xxx",
    CustomReferenceStyle,
)


@pytest.mark.sphinx(
    "text",
    testroot="footcite_roles",
    confoverrides={"bibtex_foot_reference_style": "xxx_foot_custom_xxx"},
)
def test_footcite_style_custom(app, warning) -> None:
    app.build()
    assert not warning.getvalue()
    output = (app.outdir / "index.txt").read_text(encoding="utf-8-sig")
    tests = [
        ("p", " [1] "),
        ("ps", " [1] "),
        ("t", " de Du et al[1] "),
        ("ts", " de Du & Em & Fa[1] "),
        ("ct", " De Du et al[1] "),
        ("cts", " De Du & Em & Fa[1] "),
        ("p", " [2][3] "),
        ("ps", " [2][3] "),
        ("t", " al Ap[2]; Be & Ci[3] "),
        ("ts", " al Ap[2]; Be & Ci[3] "),
        ("ct", " Al Ap[2]; Be & Ci[3] "),
        ("cts", " Al Ap[2]; Be & Ci[3] "),
        ("p", " [4][5][6] "),
        ("ps", " [4][5][6] "),
        ("t", " Ge[4]; Hu[5]; Ix[6] "),
        ("ts", " Ge[4]; Hu[5]; Ix[6] "),
        ("ct", " Ge[4]; Hu[5]; Ix[6] "),
        ("cts", " Ge[4]; Hu[5]; Ix[6] "),
    ]
    for role, text in tests:
        escaped_text = re.escape(text)
        pattern = f'":footcite:{role}:".*{escaped_text}'
        assert re.search(pattern, output) is not None


@pytest.mark.sphinx("text", testroot="footcite_roles_pre_post")
def test_footcite_pre_post(app, warning) -> None:
    app.build()
    bad_keys = ["{see}testthree", "testthree{p. 1}", "{see}testthree{p. 1}"]
    for bad_key in bad_keys:
        assert f'could not find bibtex key "{bad_key}"' in warning.getvalue()



"""
    Classes and methods to maintain any bibtex information that is stored
    outside the doctree.

    .. autoclass:: Citation
        :members:

    .. autoclass:: BibtexDomain
        :members:
"""

import ast
import re
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    cast,
)

import docutils.frontend
import docutils.nodes
import docutils.parsers.rst
import docutils.utils
import pybtex.plugin
import pybtex_docutils
import sphinx.util
from pybtex.richtext import Tag
from pybtex.style import FormattedEntry
from pybtex.style.template import FieldIsMissing
from sphinx.domains import Domain, ObjType
from sphinx.errors import ExtensionError
from sphinx.locale import _

import sphinxcontrib.bibtex.plugin

from .bibfile import BibData, normpath_filename, process_bibdata
from .citation_target import CitationTarget, parse_citation_targets
from .roles import CiteRole
from .style.referencing import BaseReferenceStyle, format_references
from .style.template import SphinxReferenceInfo

if TYPE_CHECKING:
    from pybtex.database import Entry
    from pybtex.style.formatting import BaseStyle
    from sphinx.addnodes import pending_xref
    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.environment import BuildEnvironment

    from .directives import BibliographyKey, BibliographyValue
    from .roles import CitationRef

logger = sphinx.util.logging.getLogger(__name__)


def _raise_invalid_node(node):
    """Helper method to raise an exception when an invalid node is
    visited.
    """
    raise ValueError("invalid node %s in filter expression" % node)


class _FilterVisitor(ast.NodeVisitor):

    """Visit the abstract syntax tree of a parsed filter expression."""

    entry = None
    """The bibliographic entry to which the filter must be applied."""

    cited_docnames = False
    """The documents where the entry is cited (empty if not cited)."""

    def __init__(self, entry, docname, cited_docnames):
        self.entry = entry
        self.docname = docname
        self.cited_docnames = cited_docnames

    def visit_Module(self, node):
        if len(node.body) != 1:
            raise ValueError("filter expression cannot contain multiple expressions")
        return self.visit(node.body[0])

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_BoolOp(self, node):
        outcomes = (self.visit(value) for value in node.values)
        if isinstance(node.op, ast.And):
            return all(outcomes)
        elif isinstance(node.op, ast.Or):
            return any(outcomes)
        else:  # pragma: no cover
            # there are no other boolean operators
            # so this code should never execute
            assert False, "unexpected boolean operator %s" % node.op

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.Not):
            return not self.visit(node.operand)
        else:
            _raise_invalid_node(node)

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        op = node.op
        right = self.visit(node.right)
        if isinstance(op, ast.Mod):
            # modulo operator is used for regular expression matching
            if not isinstance(left, str):
                raise ValueError("expected a string on left side of %s" % node.op)
            if not isinstance(right, str):
                raise ValueError("expected a string on right side of %s" % node.op)
            return re.search(right, left, re.IGNORECASE)
        elif isinstance(op, ast.BitOr):
            return left | right
        elif isinstance(op, ast.BitAnd):
            return left & right
        else:
            _raise_invalid_node(node)

    def visit_Compare(self, node):
        # keep it simple: binary comparators only
        if len(node.ops) != 1:
            raise ValueError("syntax for multiple comparators not supported")
        left = self.visit(node.left)
        op = node.ops[0]
        right = self.visit(node.comparators[0])
        if isinstance(op, ast.Eq):
            return left == right
        elif isinstance(op, ast.NotEq):
            return left != right
        elif isinstance(op, ast.Lt):
            return left < right
        elif isinstance(op, ast.LtE):
            return left <= right
        elif isinstance(op, ast.Gt):
            return left > right
        elif isinstance(op, ast.GtE):
            return left >= right
        elif isinstance(op, ast.In):
            return left in right
        elif isinstance(op, ast.NotIn):
            return left not in right
        else:
            # not used currently: ast.Is | ast.IsNot
            _raise_invalid_node(op)

    def visit_Name(self, node):
        """Calculate the value of the given identifier."""
        id_ = node.id
        if id_ == "type":
            return self.entry.type.lower()
        elif id_ == "key":
            return self.entry.key.lower()
        elif id_ == "cited":
            return bool(self.cited_docnames)
        elif id_ == "docname":
            return self.docname
        elif id_ == "docnames":
            return self.cited_docnames
        elif id_ == "author" or id_ == "editor":
            if id_ in self.entry.persons:
                return " and ".join(
                    str(person)  # XXX needs fix in pybtex?
                    for person in self.entry.persons[id_]
                )
            else:
                return ""
        else:
            return self.entry.fields.get(id_, "")

    def visit_Set(self, node):
        return frozenset(self.visit(elt) for elt in node.elts)

    # NameConstant is Python 3.4 only
    def visit_NameConstant(self, node):
        return node.value  # pragma: no cover

    # Constant is Python 3.6+ only
    # Since 3.8 Num, Str, Bytes, NameConstant and Ellipsis are just Constant
    def visit_Constant(self, node):
        return node.value

    # Not used on 3.8+
    def visit_Str(self, node):
        return node.s  # pragma: no cover

    def generic_visit(self, node):
        _raise_invalid_node(node)


def get_docnames(env):
    """Get document names in order."""
    rel = env.collect_relations()
    docname = (
        env.config.master_doc if sphinx.version_info < (4, 0) else env.config.root_doc
    )
    docnames = set()
    while docname is not None:
        docnames.add(docname)
        yield docname
        parent, prevdoc, nextdoc = rel[docname]
        docname = nextdoc
    for docname in sorted(env.found_docs - docnames):
        yield docname


class Citation(NamedTuple):
    """Information about a citation."""

    citation_id: str  #: Unique id of this citation.
    bibliography_key: "BibliographyKey"  #: Key of its bibliography directive.
    key: str  #: Key (with prefix).
    entry: "Entry"  #: Entry from pybtex.
    formatted_entry: "FormattedEntry"  #: Formatted entry for bibliography.
    tooltip_entry: Optional["FormattedEntry"]  #: Formatted entry for tooltip.


def env_updated(app: "Sphinx", env: "BuildEnvironment") -> Iterable[str]:
    domain = cast(BibtexDomain, env.get_domain("cite"))
    return domain.env_updated()


def parse_header(header: str, source_path: str):
    parser = docutils.parsers.rst.Parser()
    # note: types stub for docutils doesn't know about components argument
    settings = docutils.frontend.OptionParser(
        components=(docutils.parsers.rst.Parser,)  # type: ignore
    ).get_default_values()
    document = docutils.utils.new_document(source_path, settings)
    parser.parse(header, document)
    return document[0]


class BibtexDomain(Domain):
    """Sphinx domain for the bibtex extension."""

    name = "cite"
    label = "BibTeX Citations"
    data_version = 4
    initial_data = dict(
        bibdata=BibData(
            encoding="", bibfiles={}, data=pybtex.database.BibliographyData()
        ),
        bibliography_header=docutils.nodes.container(),
        bibliographies={},
        citations=[],
        citation_refs=[],
    )
    backend = pybtex_docutils.Backend()
    reference_style: BaseReferenceStyle

    @property
    def bibdata(self) -> BibData:
        """Information about the bibliography files."""
        return self.data["bibdata"]

    @property
    def bibliography_header(self) -> docutils.nodes.Element:
        return self.data["bibliography_header"]

    @property
    def bibliographies(self) -> Dict["BibliographyKey", "BibliographyValue"]:
        """Map storing information about each bibliography directive."""
        return self.data["bibliographies"]

    @property
    def citations(self) -> List[Citation]:
        """Citation data."""
        return self.data["citations"]

    @property
    def citation_refs(self) -> List["CitationRef"]:
        """Citation reference data."""
        return self.data["citation_refs"]

    def __init__(self, env: "BuildEnvironment"):
        # set up referencing style
        style = sphinxcontrib.bibtex.plugin.find_plugin(
            "sphinxcontrib.bibtex.style.referencing",
            env.app.config.bibtex_reference_style,
        )
        self.reference_style = style()
        # set up object types and roles for referencing style
        role_names = self.reference_style.role_names()
        self.object_types = dict(
            citation=ObjType(_("citation"), *role_names, searchprio=-1),
        )
        self.roles = dict((name, CiteRole()) for name in role_names)
        # initialize the domain
        super().__init__(env)
        # connect env-updated
        env.app.connect("env-updated", env_updated)
        # check config
        if env.app.config.bibtex_bibfiles is None:
            raise ExtensionError("You must configure the bibtex_bibfiles setting")
        # update bib file information in the cache
        bibfiles = [
            normpath_filename(env, "/" + bibfile)
            for bibfile in env.app.config.bibtex_bibfiles
        ]
        self.data["bibdata"] = process_bibdata(
            self.bibdata, bibfiles, env.app.config.bibtex_encoding
        )
        # parse bibliography header
        header = getattr(env.app.config, "bibtex_bibliography_header")
        if header:
            self.data["bibliography_header"] = docutils.nodes.container()
            self.data["bibliography_header"] += parse_header(
                header, "bibliography_header"
            )

    def clear_doc(self, docname: str) -> None:
        self.data["citations"] = [
            citation
            for citation in self.citations
            if citation.bibliography_key.docname != docname
        ]
        self.data["citation_refs"] = [
            ref for ref in self.citation_refs if ref.docname != docname
        ]
        for bib_key in list(self.bibliographies.keys()):
            if bib_key.docname == docname:
                del self.bibliographies[bib_key]

    def merge_domaindata(self, docnames: List[str], otherdata: Dict) -> None:
        for bib_key, bib_value in otherdata["bibliographies"].items():
            if bib_key.docname in docnames:
                self.bibliographies[bib_key] = bib_value
        for citation_ref in otherdata["citation_refs"]:
            if citation_ref.docname in docnames:
                self.citation_refs.append(citation_ref)
        # 'citations' domain data calculated in env_updated

    def env_updated(self) -> Iterable[str]:
        # This function is called when all doctrees are parsed,
        # but before any post transforms are applied. We use it to
        # determine which citations will be added to which bibliography
        # directive, and also to format the labels. We need to format
        # the labels here because they must be known when resolve_xref is
        # called.
        self.citations.clear()  # might have been restored from pickle
        docnames = list(get_docnames(self.env))
        # we keep track of this to quickly check for duplicates
        used_keys: Set[str] = set()
        used_labels: Dict[str, str] = {}
        for bibliography_key, bibliography in self.bibliographies.items():
            for entry, formatted_entry, tooltip_entry in self.get_formatted_entries(
                bibliography_key,
                docnames,
                self.env.app.config.bibtex_tooltips,
                self.env.app.config.bibtex_tooltips_style,
            ):
                key = bibliography.keyprefix + formatted_entry.key
                if bibliography.list_ == "citation" and key in used_keys:
                    logger.warning(
                        'duplicate citation for key "%s"' % key,
                        location=(bibliography_key.docname, bibliography.line),
                        type="bibtex",
                        subtype="duplicate_citation",
                    )
                self.citations.append(
                    Citation(
                        citation_id=bibliography.citation_nodes[key]["ids"][0],
                        bibliography_key=bibliography_key,
                        key=key,
                        entry=entry,
                        formatted_entry=formatted_entry,
                        tooltip_entry=tooltip_entry,
                    )
                )
                if bibliography.list_ == "citation":
                    used_keys.add(key)
                    if formatted_entry.label not in used_labels:
                        used_labels[formatted_entry.label] = key
                    elif used_labels[formatted_entry.label] != key:
                        # if used_label[label] == key then already
                        # duplicate key warning
                        logger.warning(
                            'duplicate label "%s" for keys "%s" and "%s"'
                            % (
                                formatted_entry.label,
                                used_labels[formatted_entry.label],
                                key,
                            ),
                            location=(bibliography_key.docname, bibliography.line),
                            type="bibtex",
                            subtype="duplicate_label",
                        )
        return []  # expects list of updated docnames

    def resolve_xref(
        self,
        env: "BuildEnvironment",
        fromdocname: str,
        builder: "Builder",
        typ: str,
        target: str,
        node: "pending_xref",
        contnode: docutils.nodes.Element,
    ) -> docutils.nodes.Element:
        """Replace node by list of citation references (one for each key)."""
        targets = parse_citation_targets(target)
        keys: Dict[str, CitationTarget] = {target2.key: target2 for target2 in targets}
        citations: Dict[str, Citation] = {
            cit.key: cit
            for cit in self.citations
            if cit.key in keys
            and self.bibliographies[cit.bibliography_key].list_ == "citation"
        }
        for key in keys:
            if key not in citations:
                logger.warning(
                    'could not find bibtex key "%s"' % key,
                    location=node,
                    type="bibtex",
                    subtype="key_not_found",
                )
        plaintext = pybtex.plugin.find_plugin("pybtex.backends", "plaintext")()
        references = [
            (
                citation.entry,
                citation.formatted_entry,
                SphinxReferenceInfo(
                    builder=builder,
                    fromdocname=fromdocname,
                    todocname=citation.bibliography_key.docname,
                    citation_id=citation.citation_id,
                    title=(
                        citation.tooltip_entry.text.render(plaintext).replace(
                            "\\url ", ""
                        )
                        if citation.tooltip_entry
                        else None
                    ),
                    pre_text=keys[citation.key].pre,
                    post_text=keys[citation.key].post,
                ),
            )
            for citation in citations.values()
        ]
        formatted_references = format_references(self.reference_style, typ, references)
        result_node = docutils.nodes.inline(rawsource=target)
        result_node += formatted_references.render(self.backend)
        return result_node

    def resolve_any_xref(
        self,
        env: "BuildEnvironment",
        fromdocname: str,
        builder: "Builder",
        target: str,
        node: "pending_xref",
        contnode: docutils.nodes.Element,
    ) -> List[Tuple[str, docutils.nodes.Element]]:
        """Replace node by list of citation references (one for each key),
        provided that the target has citation keys.
        """
        keys = [key.strip() for key in target.split(",")]
        citations: Set[str] = {
            cit.key
            for cit in self.citations
            if cit.key in keys
            and self.bibliographies[cit.bibliography_key].list_ == "citation"
        }
        if any(key in citations for key in keys):
            result_node = self.resolve_xref(
                env, fromdocname, builder, "p", target, node, contnode
            )
            return [("p", result_node)]
        else:
            return []

    def get_all_cited_keys(self, docnames):
        """Yield all citation keys for given *docnames* in order, then
        ordered by citation order.
        """
        for citation_ref in sorted(
            self.citation_refs, key=lambda c: docnames.index(c.docname)
        ):
            for target in citation_ref.targets:
                yield target.key

    def get_entries(self, bibfiles: List[str]) -> Iterable["Entry"]:
        """Return all bibliography entries from the bib files, unsorted (i.e.
        in order of appearance in the bib files.
        """
        for bibfile in bibfiles:
            for key in self.bibdata.bibfiles[bibfile].keys:
                yield self.bibdata.data.entries[key]

    def get_filtered_entries(
        self, bibliography_key: "BibliographyKey"
    ) -> Iterable[Tuple[str, "Entry"]]:
        """Return unsorted bibliography entries filtered by the filter
        expression.
        """
        bibliography = self.bibliographies[bibliography_key]
        for entry in self.get_entries(bibliography.bibfiles):
            key = bibliography.keyprefix + entry.key
            cited_docnames = {
                citation_ref.docname
                for citation_ref in self.citation_refs
                if key in {target.key for target in citation_ref.targets}
            }
            visitor = _FilterVisitor(
                entry=entry,
                docname=bibliography_key.docname,
                cited_docnames=cited_docnames,
            )
            try:
                success = visitor.visit(bibliography.filter_)
            except ValueError as err:
                logger.warning(
                    "syntax error in :filter: expression; %s" % err,
                    location=(bibliography_key.docname, bibliography.line),
                    type="bibtex",
                    subtype="filter_syntax_error",
                )
                # recover by falling back to the default
                success = bool(cited_docnames)
            if success or entry.key in bibliography.keys:
                yield key, entry

    def get_sorted_entries(
        self, bibliography_key: "BibliographyKey", docnames: List[str]
    ) -> Iterable[Tuple[str, "Entry"]]:
        """Return filtered bibliography entries sorted by citation order."""
        entries = dict(self.get_filtered_entries(bibliography_key))
        for key in self.get_all_cited_keys(docnames):
            try:
                entry = entries.pop(key)
            except KeyError:
                pass
            else:
                yield key, entry
        # then all remaining keys, in order of bibliography file
        for key, entry in entries.items():
            yield key, entry

    def get_formatted_entries(
        self,
        bibliography_key: "BibliographyKey",
        docnames: List[str],
        tooltips: bool,
        tooltips_style: str,
    ) -> Iterable[Tuple["Entry", "FormattedEntry", Optional["FormattedEntry"]]]:
        """Get sorted bibliography entries along with their pybtex labels,
        with additional sorting and formatting applied from the pybtex style.
        """
        bibliography = self.bibliographies[bibliography_key]
        entries = dict(self.get_sorted_entries(bibliography_key, docnames))
        style: BaseStyle = cast(
            "BaseStyle",
            pybtex.plugin.find_plugin("pybtex.style.formatting", bibliography.style)(),
        )
        style2: Optional[BaseStyle] = (
            (
                cast(
                    "BaseStyle",
                    pybtex.plugin.find_plugin(
                        "pybtex.style.formatting", tooltips_style
                    )(),
                )
                if tooltips_style
                else style
            )
            if tooltips
            else None
        )
        sorted_entries: Iterable[Entry] = style.sort(entries.values())
        labels = style.format_labels(sorted_entries)
        for label, entry in zip(labels, sorted_entries):
            try:
                yield (
                    entry,
                    style.format_entry(bibliography.labelprefix + label, entry),
                    style2.format_entry(bibliography.labelprefix + label, entry)
                    if style2
                    else None,
                )
            except FieldIsMissing as exc:
                logger.warning(
                    str(exc),
                    location=(bibliography_key.docname, bibliography.line),
                    type="bibtex",
                    subtype="missing_field",
                )
                formatted_error_entry = FormattedEntry(
                    entry.key, Tag("b", str(exc)), bibliography.labelprefix + label
                )
                yield entry, formatted_error_entry, None
